"""
Implements LoRA https://arxiv.org/abs/2106.09685 transforms on Levanter models.

LoRA is a parameter-efficient fine-tuning method that uses a low-rank factorization to reduce the number of parameters
in a model. We support LoRA using a similar approach to the one in [PEFT](https://github.com/huggingface/peft), and implement
routines to export LoRA models in a format compatible with PEFT.

To use LoRA, you need either to specify a rank, a scale (which is typically the same as rank), and a spec for the
target modules. The spec can be specified as:

- None (default), which means all linear modules
- a regex, e.g. `r".*(q_proj|k_proj|v_proj|o_proj).*"`
- a list of strings, e.g. `["q_proj", "k_proj", "v_proj", "o_proj"]`

We recommend using None, which was found to be better than the other options: https://twitter.com/Tim_Dettmers/status/1689375417189412864,
https://arxiv.org/pdf/2305.14314.pdf Section 4.

LoRA is implemented by doing "tree surgery" on the model, replacing [haliax.nn.Linear][] layers with a
[levanter.lora.LoraLinear][] layer that wraps the original linear layer.

Consider a simple model with two parameters, attention and mlp. That might look like:
   ```python
   model = Model(attention=Attention(proj_qkv=Linear), mlp=Mlp(...))
   ```
 With LoRA, our model might look more like:
   ```python
   model = Model(attention=Attention(proj_qkv=LoraLinear(wrapped=orig_proj_qkv, lora=LoraLinear(...)), mlp=Mlp(...))
   ```

 During training, you'll want to keep the base and adapter parameters separate, so that it looks like this:
   ```python
   base_model, adapter_model = partition_lora_params(model)
   base_model = Model(attention=Attention(proj_qkv=LoraLinear(wrapped=orig_proj_qkv, lora=None), mlp=Mlp(...))
   adapter_model = Model(attention=Attention(proj_qkv=LoraLinear(wrapped=None, lora=LoraLinear(...)), mlp=(...))
   ```
 and then we combine them at runtime:
   ```python
   model = combine_lora_params(base_model, lora_params=adapter_model)
   ```
 which just grounds out into a call to [equinox.combine][]
"""
import dataclasses
import functools
import json
import logging
import os
import re
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, TypeVar, Union

import equinox as eqx
import jax
from jaxtyping import PyTree

import haliax as hax
import haliax.nn as hnn
from haliax import Axis
from haliax.jax_utils import shaped_rng_split
from haliax.state_dict import (
    ModuleWithStateDictSerialization,
    StateDict,
    save_state_dict,
    to_torch_compatible_state_dict,
)

from levanter.compat.hf_checkpoints import HFCheckpointConverter, RepoRef, upload_to_hub
from levanter.logging import silence_transformer_nag
from levanter.trainer import StepInfo
from levanter.utils.cloud_utils import temp_dir_before_upload
from levanter.utils.jax_utils import join_key, key_iterator, leaf_key_paths


silence_transformer_nag()
from transformers import PreTrainedTokenizerBase  # noqa: E402


logger = logging.getLogger(__name__)


M = TypeVar("M", bound=PyTree)

# Remaining Tasks
# - bias


LORA_R = "LORA_R"


@dataclass(frozen=True)
class LoraConfig:
    target_modules: Optional[Union[List[str], str]] = None
    """modules to loraize. can either be a regex or a list of strings of module names, or None, meaning all linear modules"""
    r: int = 8  # rank of LoRA transform
    alpha: float = 8.0  # scaling factor for LoRA transform
    dropout: float = 0.0  # dropout probability for LoRA layers
    # TODO: bias

    def matches_target(self, key_path):
        if isinstance(self.target_modules, str):
            compiled = re.compile(self.target_modules)
            return compiled.match(key_path) is not None
        elif self.target_modules is None:
            return True
        else:
            return any(key_path.endswith(target) for target in self.target_modules)


def loraize(model: M, config: LoraConfig, key: jax.random.PRNGKey) -> M:
    """
    Applies LoRA transform to the given model by replacing Linear layers that match the given pattern with LoraLinear layers.
    """
    return _loraize(model, config, key, "", batch_dims=())


class LowRankLinear(eqx.Module):
    """
    A linear layer with a low-rank factorization. Used by LoRA.

    A LowRankLinear is basically 2 matrices that have a common dimension
    """

    lora_A: hnn.Linear
    lora_B: hnn.Linear
    dropout: hnn.Dropout
    scale: float = eqx.field(static=True)

    def __call__(self, x, key=None):
        if key is None and self.dropout.is_active:
            raise RuntimeError(
                "Cannot call LoraLinear with dropout and without a key if dropout is enabled."
                " The base model needs to be retrofitted to pass keys to the Linear layers."
            )
        x = self.dropout(x, key=key)
        z = self.lora_A(x)
        z = self.lora_B(z)
        return z * self.scale

    @staticmethod
    def init(In: hax.Axis, Out: Axis, r: int, alpha: float, dropout_prob: float, *, key):
        """
        Initializes a LoraLinear module.
        """
        _R = hax.Axis(LORA_R, r)
        key_A, key_B = jax.random.split(key)
        # Peft always uses out_first=True (i.e. normal Torch convention) for linear, even for gpt2-style Conv1d
        lora_A = hnn.Linear.init(In, _R, key=key_A, use_bias=False, out_first=True)
        lora_B = hnn.Linear.init(_R, Out, key=key_B, use_bias=False, out_first=True)
        dropout = hnn.Dropout(dropout_prob)

        return LowRankLinear(lora_A, lora_B, dropout, alpha / r)

    def merge(self) -> hax.NamedArray:
        return hax.dot(self.lora_A.weight, self.lora_B.weight, axis=LORA_R) * self.scale


class LoraLinear(ModuleWithStateDictSerialization):
    """
    Linear layer with LoRA transform.
    """

    wrapped: hnn.Linear
    lora: LowRankLinear

    def __call__(self, x, key=None):
        if key is not None:
            k1, k2 = jax.random.split(key)
            return self.lora(x, key=k2) + self.wrapped(x, key=k1)
        else:

            return self.lora(x) + self.wrapped(x)

    def merge(self):
        weight = self.lora.merge() + self.wrapped.weight
        return dataclasses.replace(self.wrapped, weight=weight)

    @staticmethod
    def init(wrapped: hnn.Linear, r: int, alpha: float, dropout: float = 0.0, *, key):
        """
        Initializes a LoraLinear module.
        """
        lora = LowRankLinear.init(wrapped.In, wrapped.Out, r, alpha, dropout, key=key)
        return LoraLinear(wrapped, lora)

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"wrapped": None, "lora": None}


def _is_lora_compatible_module(module):
    # TODO: more modules
    return isinstance(module, hnn.Linear)


def filter_lora_params(params: M) -> M:
    """
    Filters LoRA parameters from the given parameter tree.
    """

    return eqx.filter(params, is_lora_param, is_leaf=is_lora_param)


def partition_lora_params(params: M) -> Tuple[M, M]:
    """
    Partitions the given parameter tree into base/non-LoRA parameters and non-LoRA parameters.

    Returns:
        (base_params, lora_params)
    """
    lora_params, base_params = eqx.partition(params, is_lora_param, is_leaf=is_lora_param)
    return base_params, lora_params


def combine_lora_params(params: M, lora_params: M) -> M:
    """
    Combines the given LoRA parameters with the given parameter tree.
    """
    return eqx.combine(params, lora_params, is_leaf=is_lora_param)


def is_lora_param(node):
    return isinstance(node, LowRankLinear)


def lora_trainable_params_filter(model: M) -> M:
    """
    Creates a filter tree suitable for passing to Trainer.is_trainable marking which parameters are trainable and which
    are not.

    Returns:
       (PyTree) A filter tree marking which parameters are trainable and which are not. This filter is the same as the model,
       except every LoRA param is replaced with True and every other leaf (really, every array) is replaced with False.
    """

    # We only want to train on the lora params. The way to do this in Equinox is generally with
    # a filter tree (cf https://docs.kidger.site/equinox/examples/frozen_layer/),
    # which is a tree with the same structure (or a "tree prefix" thereof) as the model, but with
    # bools or Callable[..., bool] at the leaves. We can then pass this tree to the trainer and it
    # will only train the parameters that are True in the tree.
    # Levanter defines `is_lora_param` for this purpose, but we need to be careful about how we use it.
    # Equinox's primitives don't really have a "match all tree nodes matching a predicate" function (just
    # a "match all tree leaves matching a predicate" function), so we need to be just a bit careful.
    # Basically, we want to halt recursion in the tree whenever we hit a node that is a lora param.
    return jax.tree_util.tree_map(is_lora_param, model, is_leaf=is_lora_param)


def _loraize(model: M, config: LoraConfig, key: jax.random.PRNGKey, prefix: str, batch_dims: Tuple[Axis, ...]) -> M:
    """
    This implementation is mostly straightforward, with one major wrinkle: scan layers like Stacked, which
    add an extra batch dimension and thus require vmap, and thus require a vmap'ed LoRA transform.

    As an example, the GPT-2 Model has a Stacked[Gpt2Block] member, which means that all members of the Gpt2Block
    have an extra Layers dimension. E.g. the c_attn will have shape weight=NamedArray(['layers', 'embed', 'qkv', 'heads', 'head_size']
    even though it's defined as just Linear(In=Embed, Out=(Qkv, Heads, HeadSize)). The Stacked adds the initial Layers dimension.

    There are two ways we can approach scan layers: one is to ask implementors of lora layers to handle
    this themselves, and the other is to handle it here. The former is more flexible, but the latter is
    more convenient, even if it runs the risk of being a leaky abstraction. We choose the latter.
    """
    key_iter = key_iterator(key)

    def _is_special_module(module):
        return _is_lora_compatible_module(module) or isinstance(module, hnn.Stacked)

    def _batchify_ctor(ctor):
        # this is gross but it basically just vmaps the ctor over each batch dimension
        return functools.reduce(lambda ctor, batch_axis: hax.vmap(ctor, batch_axis), reversed(batch_dims), ctor)

    def _loraize_module(module, key_path):
        # TODO: turn into a registry
        if isinstance(module, hnn.Stacked):
            new_inner = _loraize(
                module.stacked,
                config,
                next(key_iter),
                prefix=join_key(key_path, "stacked"),
                batch_dims=batch_dims + (module.Block,),
            )
            return dataclasses.replace(module, stacked=new_inner)
        elif config.matches_target(key_path) and _is_lora_compatible_module(module):
            my_key = next(key_iter)
            batched_key = shaped_rng_split(my_key, [axis.size for axis in batch_dims])
            return _batchify_ctor(LoraLinear.init)(module, config.r, config.alpha, config.dropout, key=batched_key)
        else:
            return module

    return jax.tree_util.tree_map(
        _loraize_module,
        model,
        leaf_key_paths(model, is_leaf=_is_special_module, prefix=prefix),
        is_leaf=_is_special_module,
    )


@hax.named_jit  # needs to be inside (named) jit s.t. it works with sharded parameters
def merge_lora_modules(module: M) -> M:
    """
    Merges LoRA modules into their wrapped modules. That is, it adds the LoRA parameters to the wrapped weights,
    producing a modified base model with no LoRA parameters.
    """

    def _merge_lora_modules(module):
        if isinstance(module, LoraLinear):
            return module.merge()
        else:
            return module

    return jax.tree_util.tree_map(_merge_lora_modules, module, is_leaf=lambda node: isinstance(node, LoraLinear))


SAFETENSORS_WEIGHTS_NAME = "adapter_model.safetensors"
CONFIG_NAME = "adapter_config.json"
DEFAULT_DICT_PREFIX = "base_model.model.transformer"


def save_peft_pretrained(
    lora_model: M,
    config: LoraConfig,
    base_model_name_or_path,
    path: str,
    tokenizer: Optional[PreTrainedTokenizerBase] = None,
    *,
    prefix: Optional[str] = DEFAULT_DICT_PREFIX,
    upload_to: Optional[Union[bool, str, RepoRef]] = None,
    **upload_kwargs,
):
    """
    Saves a LoRA model as a HuggingFace checkpoint, compatible with Peft.

    Args:
        lora_model: the LoRA model to save
        path: the path to save the model to. May be a url, in which case we will use fsspec to save to that url.
        tokenizer: if provided, will save the tokenizer to the checkpoint
        prefix: the prefix to use for the LoRA parameters. Defaults to "base_model.model.transformer", which is what
            Peft seems to expect.
        upload_to: if provided, will upload the saved model to the given hf hub repo. If a string, will be interpreted
            as a repo name + branch
        upload_kwargs: kwargs to pass to the upload function
    """
    os.makedirs(path, exist_ok=True)
    hf_config = to_hf_config(config, base_model_name_or_path=base_model_name_or_path)
    state_dict = lora_state_dict(lora_model, prefix=prefix)

    with temp_dir_before_upload(path) as local_path:
        save_state_dict(state_dict, f"{local_path}/{SAFETENSORS_WEIGHTS_NAME}")
        with open(f"{local_path}/{CONFIG_NAME}", "w") as f:
            json.dump(hf_config, f)

        if tokenizer is not None:
            tokenizer.save_pretrained(local_path)

        if upload_to is True:
            upload_to = RepoRef.from_string(base_model_name_or_path)

        if upload_to:
            upload_to = RepoRef.from_string(upload_to) if isinstance(upload_to, str) else upload_to
            upload_to_hub(local_path, repo_ref=upload_to, **upload_kwargs)


def save_peft_checkpoint_callback(
    base_path,
    config: LoraConfig,
    base_model_name_or_path,
    tokenizer: Optional[PreTrainedTokenizerBase] = None,
    upload_to_hf: Optional[Union[bool, str, RepoRef]] = False,
    **hf_upload_kwargs,
):
    """
    If hf_repo is provided, this will upload the checkpoint to the huggingface hub, passing
    any additional kwargs to the huggingface_hub.upload_folder function.

    Args:
        base_path: the base path to save the checkpoint to. `/step-<step>` will be appended to this. base_path
                   may be a GCS bucket path, in which case the checkpoint will be uploaded to GCS after being written to a tmp
        config: the LoRA config to use
        base_model_name_or_path: the name or path of the base model
        tokenizer: If provided, will save the tokenizer to the checkpoint
        upload_to_hf: the repo to upload to. If a string, will be interpreted as a repo name + branch
        hf_upload_kwargs: kwargs to pass to the upload function
    """

    def cb(step: StepInfo):
        nonlocal hf_upload_kwargs
        if step.step == 0:
            return
        if upload_to_hf is not None and "commit_message" not in hf_upload_kwargs:
            my_upload_kwargs = hf_upload_kwargs.copy()
            my_upload_kwargs["commit_message"] = f"Upload for step {step.step} from Levanter"
        else:
            my_upload_kwargs = hf_upload_kwargs

        logger.info(f"Saving PEFT checkpoint for step {step.step} to {base_path}")

        save_peft_pretrained(
            step.model,
            config,
            base_model_name_or_path,
            os.path.join(base_path, f"step-{step.step}"),
            tokenizer,
            upload_to=upload_to_hf,
            **my_upload_kwargs,
        )

        logger.info("Saved checkpoint.")

    return cb


def save_merged_hf_checkpoint_callback(
    base_path,
    converter: HFCheckpointConverter,
    upload_to_hf: Optional[Union[str, RepoRef]] = None,
    **hf_upload_kwargs,
):
    """
    Saves a merged HF checkpoint for the given model. This method essentially combines the base model with the LoRA
    model using [levanter.lora.combine_lora_params][], and then saves the combined model as a HuggingFace checkpoint
    using the given converter.

    If hf_repo is provided, this will upload the checkpoint to the huggingface hub, passing any additional kwargs to the
    huggingface_hub.upload_folder function.

    Args:
        base_path: the base path to save the checkpoint to. `/step-<step>` will be appended to this. base_path
            may be a GCS bucket path, in which case the checkpoint will be uploaded to GCS after being written to a tmp
        converter: the converter to use to convert the base model to HF
        base_model: the base model to save
        upload_to_hf: the repo to upload to. If a string, will be interpreted as a repo name + branch
        **hf_upload_kwargs:
    Returns:
        callback
    """

    def save_merged_hf_model_cb(step: StepInfo):
        if step.step == 0:
            return
        if upload_to_hf is not None and "commit_message" not in hf_upload_kwargs:
            my_upload_kwargs = hf_upload_kwargs.copy()
            my_upload_kwargs["commit_message"] = f"Upload for step {step.step} from Levanter"
        else:
            my_upload_kwargs = hf_upload_kwargs

        logger.info(f"Saving merged HF model for step {step.step} to {base_path}")
        path = os.path.join(base_path, f"step-{step.step}")

        model = step.model

        save_merged_hf_model(model, converter, path, upload_to_hf=upload_to_hf, **my_upload_kwargs)

        logger.info("Saved merged checkpoint.")

    return save_merged_hf_model_cb


def save_merged_hf_model(
    lora_model: M,
    converter: HFCheckpointConverter,
    path: str,
    upload_to_hf: Optional[Union[str, RepoRef]] = None,
    **upload_kwargs,
):
    """
    Saves a merged HF checkpoint for the given model. This method essentially combines the base model with the LoRA
    model using [levanter.lora.merge_lora_modules][], and then saves the combined model as a HuggingFace checkpoint
    """
    merged_model = merge_lora_modules(lora_model)
    if upload_to_hf is None:
        upload_to_hf = False  # type: ignore
    converter.save_pretrained(
        merged_model,
        path,
        upload_to_hf=upload_to_hf,  # type: ignore
        **upload_kwargs,
    )


def to_hf_config(config: LoraConfig, base_model_name_or_path: Optional[str] = None, **kwargs) -> dict:
    """
    Converts a LoraConfig to a HuggingFace config.
    """
    # Example:
    # {
    #   "base_model_name_or_path": "decapoda-research/llama-7b-hf",
    #   "bias": "none",
    #   "enable_lora": null,
    #   "fan_in_fan_out": false,
    #   "inference_mode": true,
    #   "lora_alpha": 16,
    #   "lora_dropout": 0.05,
    #   "merge_weights": false,
    #   "modules_to_save": null,
    #   "peft_type": "LORA",
    #   "r": 16,
    #   "target_modules": [
    #     "q_proj",
    #     "k_proj",
    #     "v_proj",
    #     "o_proj"
    #   ],
    #   "task_type": "CAUSAL_LM"
    # }
    return {
        "base_model_name_or_path": base_model_name_or_path,
        "bias": "none",  # TODO: support bias
        "fan_in_fan_out": False,  # TODO: support fan_in_fan_out
        "inference_mode": True,  # TODO: support inference_mode
        "lora_alpha": config.alpha,
        "lora_dropout": 0.00,  # TODO: support dropout
        "modules_to_save": None,  # TODO: support modules_to_save?
        "peft_type": "LORA",
        "r": config.r,
        "target_modules": config.target_modules,
        "task_type": "CAUSAL_LM",  # TODO: support task_type
        **kwargs,
    }


def lora_state_dict(model: M, prefix: Optional[str] = DEFAULT_DICT_PREFIX) -> StateDict:
    """
    Returns a state dict of the LoRA parameters of the given model without other parameters.
    This method attempts to return a state dict compatible with PEFT's import method.
    """
    state_dict = to_torch_compatible_state_dict(filter_lora_params(model), prefix=prefix)
    return {k: v for k, v in state_dict.items() if v is not None}



import dataclasses
import logging
import os
import tempfile
import typing
import warnings
from contextlib import AbstractContextManager
from typing import Any, Literal, Optional

import draccus
import jax

from levanter.tracker import CompositeTracker, Tracker
from levanter.tracker.helpers import hparams_to_dict
from levanter.tracker.tensorboard import TensorboardTracker
from levanter.tracker.wandb import WandbTracker
from levanter.utils.jax_utils import is_inside_jit


logger = logging.getLogger(__name__)


_global_tracker: Optional["Tracker"] = None


def log_metrics(metrics: dict[str, Any], *, step: Optional[int], commit: Optional[bool] = None):
    """
    Log metrics to the global tracker.

    Args:
        metrics: Metrics to log
        step: Step to log at
        commit: Whether to commit the metrics. If None, uses the default for the tracker.
    """
    global _global_tracker
    if _global_tracker is None:
        raise RuntimeError("No global tracker set")

    if is_inside_jit():
        # we're inside a jit, so we need to log from the host
        if commit:
            raise ValueError("Cannot commit from inside jit")
        jit_log_metrics(metrics, step=step)
    else:
        # TODO: do we need to coerce to np here?
        _global_tracker.log(metrics, step=step)


def _no_throw_log_metrics(metrics: dict[str, Any], *, step: Optional[int], commit: Optional[bool] = None):
    try:
        if _global_tracker is None:
            warnings.warn("No global tracker set")
        else:
            _global_tracker.log(metrics, step=step, commit=False)
    except Exception:
        logger.exception("Error logging metrics")


def jit_log_metrics(metrics, *, step=None):
    """uses jax effect callback to log to wandb from the host"""
    jax.debug.callback(_no_throw_log_metrics, metrics, step=step)


def log_summary(metrics: dict[str, Any]):
    """
     Log summary metrics to the global tracker.

    Args:
         metrics: Metrics to log
    """
    global _global_tracker
    if _global_tracker is None:
        warnings.warn("No global tracker set")
        return

    _global_tracker.log_summary(metrics)


def log_hyperparameters(hparams: dict[str, Any]):
    """
     Log hyperparameters to the global tracker.

    Args:
         hparams: Hyperparameters to log
    """
    global _global_tracker
    if _global_tracker is None:
        warnings.warn("No global tracker set")
        return

    _global_tracker.log_hyperparameters(hparams)


def log_configuration(hparams: Any, config_name: Optional[str] = None):
    """
     Logs a configuration object to the global tracker. If the configuration object is a dataclass,
        it is dumped to a yaml file and logged as an artifact.

    Args:
         hparams: Hyperparameters to log
    """
    global _global_tracker
    if _global_tracker is None:
        warnings.warn("No global tracker set")
        return

    hparams_dict = hparams_to_dict(hparams)
    _global_tracker.log_hyperparameters(hparams_dict)

    if dataclasses.is_dataclass(hparams):
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = os.path.join(tmpdir, "config.yaml")
            with open(config_path, "w") as f:
                draccus.dump(hparams, f, encoding="utf-8")
                name = config_name or "config.yaml"
                _global_tracker.log_artifact(config_path, name=name, type="config")


def set_global_tracker(tracker: Tracker):
    """
    Set the global tracker. Note that setting the global tracker is not thread-safe,
    and using a tracker from multiple threads is only supported if the tracker itself is thread-safe.

    In general, it's preferred to use the context manager returned by `current_tracker` instead of this function
    except for once at the beginning of the program.

    Args:
        tracker: The tracker to set as the global tracker
        force: Whether to force setting the global tracker even if it is already set

    Examples:
        >>> from levanter.tracker import set_global_tracker, log_metrics
        >>> from levanter.tracker.wandb import WandbTracker
        >>> set_global_tracker(WandbTracker())
        >>> log_metrics({"foo": 1}, step=0)
    """
    global _global_tracker
    if _global_tracker is not None:
        warnings.warn("Global tracker is already set. Overwriting it.")
    _global_tracker = tracker


@typing.overload
def current_tracker() -> "Tracker":
    ...


@typing.overload
def current_tracker(tracker: "Tracker") -> typing.ContextManager:
    """Returns a context manager for setting the global tracker"""
    ...


def current_tracker(
    tracker: Optional[Tracker] = None,
) -> Tracker | typing.ContextManager:
    """
    Get or set the global tracker. Note that setting the global tracker is not thread-safe,
    and using a tracker from multiple threads is only supported if the tracker itself is thread-safe.

    Args:
      tracker: If provided, returns a context manager that sets the global tracker to the provided tracker when used.

    Returns:
        If no tracker is provided, returns the current global tracker.
        If a tracker is provided, returns a context manager that sets the global tracker to the provided tracker when used.

    Examples:
        >>> from levanter.tracker import current_tracker, log_metrics
        >>> from levanter.tracker.wandb import WandbTracker
        >>> with current_tracker(WandbTracker()):
        ...     log_metrics({"foo": 1}, step=0)
        ...     current_tracker().log({"foo": 2}, step=1)
    """
    global _global_tracker
    if tracker is None:
        if _global_tracker is None:
            raise RuntimeError("No global tracker set")
        return _global_tracker
    else:
        return _GlobalLoggerContextManager(tracker)


@typing.overload
def get_tracker(name: Literal["wandb"]) -> WandbTracker:
    ...


@typing.overload
def get_tracker(name: Literal["tensorboard"]) -> TensorboardTracker:
    ...


@typing.overload
def get_tracker(name: str) -> Tracker:
    ...


def get_tracker(name: str) -> Tracker:
    """
    Lookup a tracker in the current global tracker with the provided name.

    Args:
        name: Name of the tracker to lookup

    Returns:
        The tracker with the provided name

    Examples:
        >>> from levanter.tracker import get_tracker, log_metrics
        >>> from levanter.tracker.wandb import WandbTracker
        >>> with current_tracker(WandbTracker()):
        ...     log_metrics({"foo": 1}, step=0)
        ...     get_tracker("wandb").log_metrics({"foo": 2}, step=1)
    """
    tracker = current_tracker()
    if isinstance(tracker, CompositeTracker):
        for t in tracker.loggers:
            if t.name == name:
                return t
    elif tracker.name == name:
        return tracker

    raise KeyError(f"Tracker with name {name} not found")


class _GlobalLoggerContextManager(AbstractContextManager):
    def __init__(self, tracker: "Tracker"):
        self.tracker = tracker

    def __enter__(self):
        global _global_tracker
        self.old_tracker = _global_tracker
        _global_tracker = self.tracker

        return self.tracker

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _global_tracker
        _global_tracker = self.old_tracker



import levanter.checkpoint as checkpoint
import levanter.config as config
import levanter.data as data
import levanter.distributed as distributed
import levanter.eval as eval
import levanter.logging as logging
import levanter.models as models
import levanter.optim as optim
import levanter.tracker as tracker
import levanter.trainer as trainer
import levanter.visualization as visualization
from levanter.tracker import current_tracker
from levanter.trainer import initialize


__version__ = "1.2"



import tempfile

import pytest
from datasets import load_dataset
from transformers import AutoProcessor, AutoTokenizer

from levanter.data.audio import AudioDatasetSourceConfig, AudioIODatasetConfig, BatchAudioProcessor
from levanter.store.cache import SerialCacheWriter
from test_utils import skip_if_hf_model_not_accessible, skip_if_no_soundlibs


@skip_if_no_soundlibs
@skip_if_hf_model_not_accessible("openai/whisper-tiny")
def test_whisper_batch_processor():
    try:
        processor = AutoProcessor.from_pretrained("openai/whisper-tiny")
        tokenizer = AutoTokenizer.from_pretrained("openai/whisper-tiny")
        ds = load_dataset("WillHeld/test_librispeech_parquet", split="validation").select_columns(["audio", "text"])
        batch_processor = BatchAudioProcessor(processor, tokenizer)
        inputs = [
            (audio["array"], audio["sampling_rate"], text) for audio, text in zip(ds[:16]["audio"], ds[:16]["text"])
        ]
        batch_processor(inputs)
    except FileNotFoundError:
        pytest.skip("No whisper model found. Probably HF is being flaky.")


@skip_if_no_soundlibs
@skip_if_hf_model_not_accessible("openai/whisper-tiny")
def test_hf_audio_loading():
    # Use the Real Librispeech Valudation. Testing one doesn't support streaming.
    ac = AudioDatasetSourceConfig(id="WillHeld/test_librispeech_parquet", text_key="text")
    audio_iterator = ac.doc_iterator("validation")
    for i in range(10):
        audio, sample, text = next(audio_iterator)


@skip_if_no_soundlibs
@skip_if_hf_model_not_accessible("openai/whisper-tiny")
def test_hf_audio_loading_source():
    # Use the Real Librispeech Valudation. Testing one doesn't support streaming.
    ac = AudioDatasetSourceConfig(id="WillHeld/test_librispeech_parquet", text_key="text")
    audio_iterator = iter(ac.get_shard_source("validation"))
    for i in range(10):
        audio, sample, text = next(audio_iterator)


@skip_if_no_soundlibs
@skip_if_hf_model_not_accessible("openai/whisper-tiny")
@pytest.mark.asyncio
async def test_hf_audio_ray_pipeline():
    # Use the Real Librispeech Valudation. Testing one doesn't support streaming.
    with tempfile.TemporaryDirectory() as tmpdir:
        ac = AudioIODatasetConfig(
            cache_dir=str(tmpdir), id="WillHeld/test_librispeech_parquet", text_key="text", max_length=1024
        )
        validation = ac.validation_set()
        for i in range(10):
            t = (await validation.get_batch([i]))[0]
            assert t["input_features"].shape == (80, 3000), t["input_features"].shape
            assert t["input_ids"].shape == (1024,), t["input_ids"].shape
            assert t["attention_mask"].shape == (1024,), t["attention_mask"].shape


@skip_if_no_soundlibs
@skip_if_hf_model_not_accessible("openai/whisper-tiny")
def test_hf_audio_serial_cache():
    # Use the Real Librispeech Valudation. Testing one doesn't support streaming.
    ac = AudioIODatasetConfig(id="WillHeld/test_librispeech_parquet", text_key="text")

    processor = AutoProcessor.from_pretrained("openai/whisper-tiny")
    tokenizer = AutoTokenizer.from_pretrained("openai/whisper-tiny")
    batch_processor = BatchAudioProcessor(processor, tokenizer, max_length=1024)

    with tempfile.TemporaryDirectory() as tmpdir:
        with SerialCacheWriter(tmpdir, batch_processor.output_exemplar) as writer:
            for i, ex in enumerate(ac.get_shard_source("validation")):
                writer.write_batch(batch_processor([ex]))
                if i > 10:
                    break

        cache = writer.result()

        for ex in cache.get_batch_sync(list(range(10))):
            assert ex["input_features"].shape == (80, 3000), ex["input_features"].shape
            assert ex["input_ids"].shape == (1024,), ex["input_ids"].shape
            assert ex["attention_mask"].shape == (1024,), ex["attention_mask"].shape


@skip_if_no_soundlibs
@skip_if_hf_model_not_accessible("openai/whisper-tiny")
def test_metadata_works():
    processor = AutoProcessor.from_pretrained("openai/whisper-tiny")
    tokenizer = AutoTokenizer.from_pretrained("openai/whisper-tiny")
    batch_processor = BatchAudioProcessor(processor, tokenizer)
    # test this doesn't throw
    assert len(batch_processor.metadata)



import asyncio

import jax.random
import pytest

from levanter.data import EraShufflingDataset, PermutationDataset
from levanter.data.dataset import ListAsyncDataset


@pytest.mark.asyncio
async def test_length_of_sequence_dataset_is_accurate():
    data = [1, 2, 3]
    dataset = ListAsyncDataset(data)
    assert (await dataset.current_len()) == 3
    assert not (await dataset.final_length_is_known())
    dataset.finalize()
    assert (await dataset.current_len()) == 3
    assert await dataset.final_length_is_known()
    assert (await dataset.async_len()) == 3


@pytest.mark.asyncio
async def test_list_dataset_get_item_returns_correct_item():
    data = ["a", "b", "c"]
    dataset = ListAsyncDataset(data)
    assert await dataset.getitem_async(1) == "b"


@pytest.mark.asyncio
async def test_list_async_dataset_appends_and_finalizes_correctly():
    dataset = ListAsyncDataset([])
    dataset.append("a")
    dataset.finalize()
    assert await dataset.async_len() == 1
    assert await dataset.get_batch([0]) == ["a"]


@pytest.mark.asyncio
async def test_permutation_dataset_is_at_least_sometimes_permuted():
    for seed in range(10):
        data = [1, 2, 3, 4]
        dataset = ListAsyncDataset(data, is_complete=True)
        permuted_dataset = PermutationDataset(dataset, jax.random.PRNGKey(seed))
        if await permuted_dataset.get_batch([0, 1, 2, 3]) != [1, 2, 3, 4]:
            return

    pytest.fail("PermutationDataset did not permute the data")


@pytest.mark.asyncio
async def test_era_shuffling_dataset_returns_correct_length():
    data = list(range(100))
    dataset = ListAsyncDataset(data, is_complete=False)
    era_length = 10
    key = jax.random.PRNGKey(0)
    shuffling_dataset = EraShufflingDataset(dataset, era_length, key=key)
    assert await shuffling_dataset.current_len() == 100
    assert not await shuffling_dataset.final_length_is_known()

    dataset.append(1)
    assert await shuffling_dataset.current_len() == 100


@pytest.mark.asyncio
async def test_era_shuffling_dataset_get_batch_returns_shuffled_batch():
    data = list(range(20))
    dataset = ListAsyncDataset(data)
    dataset.finalize()
    era_length = 5
    key = jax.random.PRNGKey(0)
    shuffling_dataset = EraShufflingDataset(dataset, era_length, key=key)
    batch_indices = [0, 1, 2, 3, 4]
    batch = await shuffling_dataset.get_batch(batch_indices)
    assert set(batch) == set([0, 1, 2, 3, 4])  # Ensures all elements are from the first era but does not assume order
    assert batch != [0, 1, 2, 3, 4]  # Ensures the batch is shuffled


@pytest.mark.asyncio
async def test_era_shuffling_can_grow():
    data = list(range(5))
    dataset = ListAsyncDataset(data)
    era_length = 5
    key = jax.random.PRNGKey(0)
    shuffling_dataset = EraShufflingDataset(dataset, era_length, key=key)
    batch_indices = [0, 1, 2, 3, 4]
    batch = await shuffling_dataset.get_batch(batch_indices)
    assert set(batch) == set([0, 1, 2, 3, 4])

    for i in range(5):
        dataset.append(i + 5)

    assert await shuffling_dataset.current_len() == 10
    assert not await shuffling_dataset.final_length_is_known()
    batch = await shuffling_dataset.get_batch(list(range(10)))

    assert set(batch) == set(range(10))
    assert set(batch[0:5]) == set([0, 1, 2, 3, 4])
    assert set(batch[5:10]) == set([5, 6, 7, 8, 9])

    # now make sure that we can await data and it does get fulfilled
    # this should timeout if we try to await it
    coro = dataset.get_batch([11])
    try:
        await asyncio.wait_for(coro, timeout=0.1)
        pytest.fail("Should have timed out")
    except asyncio.TimeoutError:
        pass

    async def append_data():
        await asyncio.sleep(0.1)
        for i in range(10, 15):
            dataset.append(i)

    coro = dataset.getitem_async(11)

    _, r = await asyncio.gather(append_data(), coro)
    assert r in range(10, 15)

    coro2 = shuffling_dataset.wait_until_len_at_least(20)
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(coro2, timeout=0.1)

    assert await shuffling_dataset.current_len() == 15

    coro2 = shuffling_dataset.wait_until_len_at_least(20)
    dataset.append(15)
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(coro2, timeout=0.1)

    assert await shuffling_dataset.current_len() == 15

    coro2 = shuffling_dataset.wait_until_len_at_least(20)
    dataset.finalize()
    await asyncio.wait_for(coro2, timeout=0.1)

    assert await dataset.async_len() == 16
    assert await shuffling_dataset.current_len() == 16

    coro = shuffling_dataset.get_batch(list(range(16)))

    batch = await coro
    assert set(batch) == set(range(16))



import time
from typing import Iterator

import pytest
import ray

from levanter.store._prefetch_actor import RayPrefetchQueue


def _sleep_until(condition, timeout=5, message="Condition not met within timeout"):
    start = time.time()
    while not condition():
        if time.time() - start > timeout:
            pytest.fail(message)
        time.sleep(0.1)


@pytest.fixture(scope="module", autouse=True)
def ray_init_and_shutdown():
    ray.init("local", ignore_reinit_error=True)
    yield
    ray.shutdown()


@pytest.mark.ray
def test_initialization_and_basic_functionality():
    def simple_producer():
        for i in range(10):
            yield i

    actor = RayPrefetchQueue(simple_producer)
    results = [actor.get_next() for _ in range(10)]
    assert results == list(range(10))


@pytest.mark.ray
def test_queue_size_limit():
    def simple_producer() -> Iterator[ray.ObjectRef]:
        for i in range(100):
            yield i

    actor = RayPrefetchQueue(simple_producer, max_queue_size=10)
    # Allow some time for the queue to fill up
    _sleep_until(lambda: actor.queue_size() == 10)

    # get a few items to make some space
    [actor.get_next() for _ in range(5)]
    _sleep_until(lambda: actor.queue_size() == 10, message="Queue size did not reach 10")


@pytest.mark.ray
def test_stop_functionality():
    def simple_producer():
        for i in range(10000):
            yield i

    actor = RayPrefetchQueue(simple_producer)
    actor.stop()

    _sleep_until(lambda: actor.is_stopped(), message="Actor did not stop")


@pytest.mark.ray
def test_exception_handling():
    def faulty_producer():
        for i in range(5):
            yield i
        raise ValueError("Test exception")

    actor = RayPrefetchQueue(faulty_producer)
    results = []
    try:
        for _ in range(10):
            results.append(actor.get_next())
    except ValueError as e:
        assert "Test exception" in str(e)  # Ray puts a lot of crap in the exception message
    assert results == list(range(5))


@pytest.mark.ray
def test_empty_producer():
    def empty_producer() -> Iterator[ray.ObjectRef]:
        if False:
            yield

    actor = RayPrefetchQueue(empty_producer)
    with pytest.raises(StopIteration):
        actor.get_next()


@pytest.mark.ray
def test_multiple_consumers():
    def simple_producer() -> Iterator[ray.ObjectRef]:
        for i in range(20):
            yield i

    actor = RayPrefetchQueue(simple_producer)
    results = [actor.get_next() for _ in range(10)]
    results += [actor.get_next() for _ in range(10)]
    assert results == list(range(20))


@pytest.mark.ray
def test_producer_completion():
    def simple_producer():
        for i in range(10):
            yield i

    actor = RayPrefetchQueue(simple_producer)
    results = []
    try:
        while True:
            results.append(actor.get_next())
    except StopIteration:
        pass
    assert results == list(range(10))


@pytest.mark.ray
def test_drain_queue():
    def simple_producer():
        for i in range(10):
            yield i

    actor = RayPrefetchQueue(simple_producer)

    all_results = []

    for tot in range(0, 5):
        out = actor.drain_available(tot)
        assert len(out) <= tot
        all_results.extend(out)

    while len(all_results) < 10:
        all_results.append(actor.get_next())

    assert all_results == list(range(10))



import tempfile

import jax.numpy as jnp
from transformers import AutoTokenizer

import haliax as hax

from levanter.data.text import BatchTokenizer, LMDatasetConfig
from levanter.models.lm_model import LmExample
from levanter.models.loss import next_token_loss
from tests.test_utils import skip_if_hf_model_not_accessible


def test_dont_blow_up_without_validation_set():
    with tempfile.TemporaryDirectory() as tmpdir:
        config = LMDatasetConfig(
            train_urls=["kaa"],
            validation_urls=[],
            cache_dir=tmpdir,
        )

        # mostly just making sure this doesn't blow up
        assert config.validation_set(10) is None


def test_lm_example_handles_ignore_id():
    Pos = hax.Axis("Pos", 10)
    Vocab = hax.Axis("vocab", Pos.size + 1)
    Embed = hax.Axis("embed", 10)
    tokens = hax.arange(Pos, dtype=jnp.int32)

    ignore_id = 6

    ex_ignore = LmExample.causal(tokens, ignore_id=ignore_id)
    ex_no_ignore = LmExample.causal(tokens)
    assert ex_ignore.loss_mask[Pos, ignore_id - 1] == 0

    logits = hax.ones((Pos, Embed))
    lm_head = hax.zeros((Embed, Vocab))
    lm_head = lm_head.at[Vocab, ignore_id].set(-100)

    ignored_loss = next_token_loss(Pos, Embed, Vocab, logits, lm_head, tokens, loss_mask=ex_ignore.loss_mask)
    no_ignore_loss = next_token_loss(Pos, Embed, Vocab, logits, lm_head, tokens, loss_mask=ex_no_ignore.loss_mask)

    assert no_ignore_loss.item() >= ignored_loss.item() + 100 / Pos.size


def test_merge_split_encodings():
    tokenizer = AutoTokenizer.from_pretrained("gpt2")
    # make this very short for testing

    lorem = """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."""

    short_batch_tokenizer = BatchTokenizer(tokenizer, _workaround_len=len(lorem) // 3)
    # force this
    short_batch_tokenizer._needs_long_sequence_workaround = True

    batch_tokenizer = BatchTokenizer(tokenizer, _workaround_len=50000)
    batch = [lorem]

    short_out = short_batch_tokenizer(batch)
    reg_out = batch_tokenizer(batch)

    assert short_out == reg_out


@skip_if_hf_model_not_accessible("meta-llama/Llama-2-7b-hf")
def test_llama_tokenizer_needs_long_sequence_workaround():
    tokenizer = AutoTokenizer.from_pretrained("meta-llama/Llama-2-7b-hf")
    batch_tokenizer = BatchTokenizer(tokenizer)
    assert batch_tokenizer._needs_long_sequence_workaround



import os
import tempfile

import jax
import pytest

import levanter.main.train_lm as train_lm
import tiny_test_corpus
from levanter.distributed import RayConfig
from levanter.tracker.wandb import WandbConfig


@pytest.mark.entry
def test_train_lm():
    # just testing if train_lm has a pulse
    with tempfile.TemporaryDirectory() as tmpdir:
        data_config, _ = tiny_test_corpus.construct_small_data_cache(tmpdir)
        try:
            config = train_lm.TrainLmConfig(
                data=data_config,
                model=train_lm.Gpt2Config(
                    num_layers=2,
                    num_heads=2,
                    seq_len=64,
                    hidden_dim=32,
                    attn_backend=None,  # use default for platform
                ),
                trainer=train_lm.TrainerConfig(
                    num_train_steps=2,
                    train_batch_size=len(jax.devices()),
                    max_eval_batches=1,
                    wandb=WandbConfig(mode="disabled"),
                    require_accelerator=False,
                    ray=RayConfig(auto_start_cluster=False),
                ),
            )
            train_lm.main(config)
        finally:
            try:
                os.unlink("wandb")
            except Exception:
                pass



from levanter.utils.py_utils import actual_sizeof


def test_actual_sizeof():
    d1 = {"a": 1, "b": 2}
    d2 = {"a": "this is a string", "b": "this is another string"}

    assert actual_sizeof(d1) < actual_sizeof(d2)



import os
import re
from typing import TypeAlias

from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast

from levanter.logging import silence_transformer_nag
from levanter.utils.py_utils import logical_cpu_core_count


silence_transformer_nag()

_HF_TOKENIZER_OFF_VALUES = {"off", "false", "f", "no", "n", "0"}

HfTokenizer: TypeAlias = PreTrainedTokenizerFast | PreTrainedTokenizer
"""
Type alias for a Hugging Face tokenizer. This is a union of the two tokenizer types.
While there is PreTrainedTokenizerBase, it doesn't have all methods that are implemented in both
PreTrainedTokenizer and PreTrainedTokenizerFast. grumble grumble.
"""


def num_cpus_used_by_tokenizer(tokenizer) -> int:
    if getattr(tokenizer, "is_fast", False):
        if os.getenv("TOKENIZERS_PARALLELISM", "true").lower() in _HF_TOKENIZER_OFF_VALUES:
            return 1
        else:
            # This is a bit hacky, but HF's fast tokenizers are parallelized under the hood.
            # we reserve a couple of cores just so Ray has somewhere to run the coordinator.
            # Empirically it doesn't usually exceed 16-20, and it's useful to have some slack
            return min(max(1, logical_cpu_core_count() - 2), 12)
    else:
        return 1


def byte_length_of_token(tokenizer, idx: int) -> int:
    # this is a pain because we want the prefix spaces, but we don't want extra noise for bytes
    # e.g. in llama
    # >>> t.convert_ids_to_tokens(q[2])
    # '▁this'
    # >>> t.convert_ids_to_tokens(25)
    # '<0x16>'
    # We want the _ (as a single byte, not the 3 it's encoded as) but not the <0x16>, which should instead be a single byte \x16
    # decode strips the prefix spaces, but does correctly handle the <0x16> case
    # we can avoid prefix space issues by prepending another token before decoding, then stripping
    repr = tokenizer.convert_ids_to_tokens(idx)
    if idx in tokenizer.all_special_ids:
        # NB: special tokens don't have bytes, but they contribute to perplexity/bits
        return 0
    # handle bytes specially. This is a bit of a hack, but there's no other way
    elif m := re.match(r"<0x([0-9A-Fa-f]+)>", repr):
        return len(bytes.fromhex(m.group(1)))
    else:
        extra_token = tokenizer(".", add_special_tokens=False)["input_ids"][0]
        excess_bytes = len(".".encode("utf-8"))
        decoded = tokenizer.decode([extra_token, idx]).encode("utf-8")
        return len(decoded) - excess_bytes



import asyncio
import os
from typing import Generic, List, Sequence, TypeVar

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
from jaxtyping import PyTree

from haliax.jax_utils import is_jax_array_like

from .jagged_array import JaggedArrayStore, PreparedBatch


T = TypeVar("T", bound=PyTree)


# TODO at some point if we turn this into a real library, it would be nice to store the schema
# TODO: some data is probably best not stored as a jagged array, but as a flat array?
# TODO: also sometimes we might want a rowstore actually


def heuristic_is_leaf(x):
    if isinstance(x, list):
        return jnp.isscalar(x[0])
    else:
        return False


def heuristic_is_leaf_batched(x):
    if isinstance(x, list):
        return jnp.isscalar(x[0]) or is_jax_array_like(x[0])
    else:
        return False


class TreeStore(Generic[T]):
    """
    A TreeStoreBuilder stores batched data as a tree of ragged arrays.
    """

    path: str
    mode: str
    tree: PyTree[JaggedArrayStore]

    def __init__(self, tree, path: str, mode: str):
        self.path = path
        self.mode = mode
        self.tree = tree

    @property
    def batch_preparer(self):
        return TreeBatchPreparer(jtu.tree_map(lambda writer: 9, self.tree, is_leaf=heuristic_is_leaf))

    @staticmethod
    def open(exemplar: T, path: str, *, mode="a", cache_metadata: bool = False) -> "TreeStore":
        """
        Open a TreeStoreBuilder from a file.
        """
        tree = _construct_builder_tree(exemplar, path, mode, cache_metadata)
        return TreeStore(tree, path, mode)

    def append(self, ex: T):
        return self.extend([ex])

    def extend(self, batch: Sequence[T]):
        """
        Append a batch of data to the store.
        """
        jtu.tree_map(
            lambda writer, *xs: writer.extend([np.asarray(x) for x in xs]),
            self.tree,
            *batch,
            is_leaf=heuristic_is_leaf,
        )

    def extend_with_batch(self, batch: T):
        """
        Append a batch of data (as a pytree with batched leaves) to the store.

        This method works only when the "leaves" are lists of numpy arrays or scalars.
        For instance, HF's BatchEncoding is a dict of lists of numpy arrays.
        """
        jtu.tree_map(
            lambda writer, xs: writer.extend(xs if isinstance(xs, PreparedBatch) else [np.asarray(x) for x in xs]),
            self.tree,
            batch,
            is_leaf=heuristic_is_leaf_batched,
        )

    async def extend_with_batch_async(self, batch: T):
        """
        Append a batch of data (as a pytree with batched leaves) to the store.

        This method works only when the "leaves" are lists of numpy arrays or scalars.
        For instance, HF's BatchEncoding is a dict of lists of numpy arrays.
        """
        futures = jtu.tree_map(
            lambda writer, xs: writer.extend_async(
                xs if isinstance(xs, PreparedBatch) else [np.asarray(x) for x in xs]
            ),
            self.tree,
            batch,
            is_leaf=heuristic_is_leaf_batched,
        )

        await asyncio.gather(*jax.tree_leaves(futures))

    def trim_to_size(self, size: int):
        """
        Trim the store to a given size.
        """
        # TODO These all return ts Futures so in theory we could await them all at once
        jtu.tree_map(lambda writer: writer.trim_to_size(size), self.tree, is_leaf=heuristic_is_leaf)

    async def trim_to_size_async(self, size: int):
        """
        Trim the store to a given size.
        """
        futures = jtu.tree_map(lambda writer: writer.trim_to_size_async(size), self.tree, is_leaf=heuristic_is_leaf)
        leaves, structure = jax.tree_flatten(futures)

        await asyncio.gather(*leaves)

    def reload(self) -> "TreeStore":
        """
        Close the builder and return a TreeStore.
        """
        tree = jtu.tree_map(lambda builder: builder.reload(), self.tree, is_leaf=heuristic_is_leaf)
        return TreeStore(tree, self.path, self.mode)

    def __len__(self):
        if self.tree is None:
            return 0
        else:
            return len(jax.tree.leaves(self.tree)[0])

    async def get_batch(self, indices) -> List[T]:
        grouped = jtu.tree_map(lambda reader: reader.get_batch(indices), self.tree, is_leaf=heuristic_is_leaf)

        leaves, structure = jtu.tree_flatten(grouped, is_leaf=heuristic_is_leaf)

        awaited_leaves = await asyncio.gather(*leaves)
        return [jtu.tree_unflatten(structure, [leaf[i] for leaf in awaited_leaves]) for i in range(len(indices))]

    def __getitem__(self, item):
        if self.tree is None:
            raise IndexError("No data in store")
        elif isinstance(item, slice):
            # debatch
            leaves, structure = jax.tree.flatten(self.tree, is_leaf=heuristic_is_leaf)
            # batched_items = jtu.tree_map(lambda reader: reader[item], self.tree, is_leaf=heuristic_is_leaf)
            batched_item_leaves = [leaf[item] for leaf in leaves]
            num_items = len(leaves[0])
            return [jtu.tree_unflatten(structure, [leaf[i] for leaf in batched_item_leaves]) for i in range(num_items)]
        else:
            return jtu.tree_map(lambda reader: reader[item], self.tree, is_leaf=heuristic_is_leaf)

    def __iter__(self):
        if self.tree is None:
            return
        else:
            for i in range(len(self)):
                yield self[i]

    def get_batch_sync(self, indices) -> List[T]:
        # TODO: would be better to batch these up
        grouped = jtu.tree_map(lambda reader: reader.get_batch_sync(indices), self.tree, is_leaf=heuristic_is_leaf)

        out = [jtu.tree_map(lambda _, leaf: leaf[i], self.tree, grouped) for i in range(len(indices))]

        return out

    async def async_len(self) -> int:
        return await jax.tree.leaves(self.tree)[0].num_rows_async()


def _construct_builder_tree(exemplar, path, mode, cache_metadata):
    def open_builder(tree_path, item):
        item = np.asarray(item)
        rank = item.ndim
        render_tree_path = "/".join(_render_path_elem(x) for x in tree_path)
        return JaggedArrayStore.open(
            os.path.join(path, render_tree_path),
            mode=mode,
            item_rank=rank,
            dtype=item.dtype,
            cache_metadata=cache_metadata,
        )

    return jtu.tree_map_with_path(open_builder, exemplar, is_leaf=heuristic_is_leaf)


def _render_path_elem(x):
    match x:
        case jtu.DictKey(key):
            return f"{key}"
        case jtu.GetAttrKey(key):
            return f"{key}"
        case jtu.SequenceKey(i):
            return f"{i}"
        case jtu.FlattenedIndexKey(i):
            return f"{i}"
        case _:
            return str(x)


class TreeBatchPreparer(Generic[T]):
    def __init__(self, exemplar: T):
        self.exemplar = exemplar

    def __call__(self, batch: List[T]) -> PyTree:
        return jtu.tree_map(
            lambda _, *xs: PreparedBatch.from_batch([np.asarray(x) for x in xs]),
            self.exemplar,
            *batch,
            is_leaf=heuristic_is_leaf,
        )



import logging
from dataclasses import dataclass
from functools import cached_property
from typing import Optional

import equinox as eqx
import jax
from jax.sharding import Mesh

from haliax import Axis

import levanter
from levanter.checkpoint import load_checkpoint
from levanter.compat.hf_checkpoints import RepoRef, load_tokenizer
from levanter.models.gpt2 import Gpt2Config
from levanter.models.lm_model import LmConfig, LmHeadModel
from levanter.utils.jax_utils import is_inexact_arrayish, use_cpu_device


logger = logging.getLogger(__name__)


@dataclass
class ConvertLmConfig:
    checkpoint_path: str
    output_dir: str
    upload_to_hf: Optional[RepoRef] = None  # if specified, attempt to upload this checkpoint to the hf hub

    model: LmConfig = Gpt2Config()
    save_tokenizer: bool = True  # if True, save the tokenizer to the output directory
    tokenizer: str = "gpt2"
    override_vocab_size: Optional[int] = None  # if specified, override the vocab size in the config

    config_overrides: Optional[dict] = None  # if specified, override the config with these values

    @cached_property
    def the_tokenizer(self):
        return load_tokenizer(self.tokenizer)


def main(config: ConvertLmConfig):
    logger.setLevel(logging.INFO)
    tokenizer = config.the_tokenizer

    vocab_size = config.override_vocab_size or len(tokenizer)
    Vocab = Axis("vocab", vocab_size)

    key = jax.random.PRNGKey(0)

    with use_cpu_device(), Mesh([jax.local_devices(backend="cpu")[0]], "dev"):
        model: LmHeadModel = eqx.filter_eval_shape(config.model.build, Vocab, key=key)
        trainable, non_trainable = eqx.partition(model, is_inexact_arrayish)
        # TODO: don't load the entire checkpoint into CPU memory when we only need our share of the model
        trainable = load_checkpoint(trainable, config.checkpoint_path, subpath="model")

        assert trainable is not None
        model = eqx.combine(trainable, non_trainable)

        if config.override_vocab_size:
            model = model.resize_vocab(config.override_vocab_size)

        converter = model.config.hf_checkpoint_converter().replaced(tokenizer=tokenizer)

        converter.save_pretrained(
            model,
            config.output_dir,
            upload_to_hf=config.upload_to_hf or False,
            save_tokenizer=config.save_tokenizer,
        )


if __name__ == "__main__":
    levanter.config.main(main)()



from ._preprocessor import BatchProcessor
from .dataset import AsyncDataset, ListAsyncDataset, MappedAsyncDataset, SyncDataset
from .loader import DataLoader
from .mixture import MixtureDataset, StopStrategy
from .permutation import EraShufflingDataset, PermutationDataset
from .sharded_datasource import ShardedDataSource, datasource_from_hf, datasource_from_json, datasource_from_jsonl
from .utils import batched


__all__ = [
    "batched",
    "ShardedDataSource",
    "datasource_from_hf",
    "datasource_from_jsonl",
    "datasource_from_json",
    "BatchProcessor",
    "AsyncDataset",
    "MappedAsyncDataset",
    "SyncDataset",
    "ListAsyncDataset",
    "DataLoader",
    "MixtureDataset",
    "StopStrategy",
]



import tempfile

import equinox as eqx
import jax
import numpy as np
import pytest
import transformers
from jax import random

import haliax as hax

from levanter.models.attention import AttentionMask
from levanter.models.mistral import MistralConfig, MistralLMHeadModel
from test_utils import check_load_config, check_model_works_with_seqlen, parameterize_with_configs, skip_if_no_torch


@skip_if_no_torch
def test_mistral_config():
    # load HF config and convert to levanter config
    hf_config = transformers.MistralConfig.from_pretrained("mistralai/Mistral-7B-v0.1")
    mistral_config = MistralConfig.from_hf_config(hf_config)

    # convert back to HF config
    config_overrides = {
        "_name_or_path": hf_config._name_or_path,
        "architectures": hf_config.architectures,
        "torch_dtype": hf_config.torch_dtype,
    }
    new_hf_config = mistral_config.to_hf_config(
        vocab_size=hf_config.vocab_size,
        config_overrides=config_overrides,
    )

    # assert the content in new_hf_config is the same as hf_config
    for k in new_hf_config.__dict__.keys():
        if k in ["_commit_hash", "transformers_version"]:
            continue
        assert getattr(new_hf_config, k) == getattr(
            hf_config, k
        ), f"{k} {getattr(new_hf_config, k)} != {getattr(hf_config, k)}"


@pytest.mark.parametrize("num_kv_heads", [1, 2, 4])
def test_mistral_lm_head_model(num_kv_heads):
    mistral_config = _get_mistral_config(num_kv_heads=num_kv_heads)
    Batch = hax.Axis("batch", 2)
    Vocab = hax.Axis("vocab", 1000)
    Pos = mistral_config.Pos
    input_ids = hax.random.randint(random.PRNGKey(0), (Batch, Pos), 0, Vocab.size)
    mask = AttentionMask.causal()

    def fn(input_ids, mask):
        return MistralLMHeadModel.init(Vocab=Vocab, config=mistral_config, key=random.PRNGKey(0))(input_ids, mask)

    out = eqx.filter_eval_shape(fn, input_ids, mask)
    assert out.array.shape == (Batch.size, Pos.size, Vocab.size)


@pytest.mark.parametrize("use_flash", [True, False])
@pytest.mark.parametrize("num_kv_heads", [1, 2, 4])
def test_mistral_lm_head_model_bwd(use_flash, num_kv_heads):
    llama_config = _get_mistral_config(use_flash=use_flash, num_kv_heads=num_kv_heads)
    Batch = hax.Axis("batch", 2)
    Vocab = hax.Axis("vocab", 1000)
    Pos = llama_config.Pos
    input_ids = hax.random.randint(random.PRNGKey(0), (Batch, Pos), 0, Vocab.size)
    mask = AttentionMask.causal()

    llama_model = MistralLMHeadModel.init(Vocab=Vocab, config=llama_config, key=random.PRNGKey(0))

    def f(llama_model, input_ids, mask):
        out = llama_model(input_ids, mask)
        return hax.sum(out).scalar()

    _, grads = eqx.filter_eval_shape(eqx.filter_value_and_grad(f), llama_model, input_ids, mask)


@skip_if_no_torch
@pytest.mark.parametrize("num_kv_heads", [1, 2, 4])
def test_mistral_roundtrip(num_kv_heads):
    import torch
    from transformers import AutoModelForCausalLM, MistralForCausalLM

    config = MistralConfig(
        seq_len=128,
        hidden_dim=16,
        num_heads=4,
        num_kv_heads=num_kv_heads,
        gradient_checkpointing=False,
    )
    converter = config.hf_checkpoint_converter()

    Vocab = hax.Axis("vocab", 1000)
    hf_config = config.to_hf_config(Vocab.size)

    # Make input and attn_mask
    input = hax.random.randint(random.PRNGKey(0), config.Pos, 0, Vocab.size)
    attn_mask = AttentionMask.causal()
    input_torch = torch.from_numpy(np.array(input.array)).to(torch.int32).unsqueeze(0)

    torch.random.manual_seed(0)

    torch_model = MistralForCausalLM(hf_config)
    torch_model.eval()

    torch_out = torch_model(input_torch)
    torch_out = torch_out.logits[0].detach().cpu().numpy()
    torch_out = jax.nn.softmax(torch_out, axis=-1)

    with tempfile.TemporaryDirectory() as tmpdir:
        torch_model.save_pretrained(f"{tmpdir}/torch_model")

        model = converter.load_pretrained(
            converter.default_config.model_type, ref=f"{tmpdir}/torch_model", resize_vocab_to_match_tokenizer=False
        )

        def compute(input):
            model_output = model(input, attn_mask=attn_mask)
            return hax.nn.softmax(model_output, axis=model.Vocab)

        compute = jax.jit(compute)
        jax_out = compute(input).array

        assert torch_out.shape == jax_out.shape, f"{torch_out.shape} != {jax_out.shape}"
        assert np.isclose(torch_out, np.array(jax_out), rtol=1e-4, atol=1e-4).all(), f"{torch_out} != {jax_out}"

        converter.save_pretrained(model, f"{tmpdir}/lev_model", save_reference_code=False)
        torch_model2 = AutoModelForCausalLM.from_pretrained(f"{tmpdir}/lev_model")
        torch_model2.eval()

        torch_out2 = torch_model2(input_torch)
        torch_out2 = torch_out2.logits[0].detach().cpu().numpy()
        torch_out2 = jax.nn.softmax(torch_out2, axis=-1)
        assert torch_out2.shape == jax_out.shape, f"{torch_out2.shape} != {jax_out.shape}"
        assert np.isclose(torch_out2, np.array(jax_out), rtol=1e-4, atol=1e-4).all(), f"{torch_out2} != {jax_out}"


def _get_mistral_config(use_flash=False, num_kv_heads=4) -> MistralConfig:
    return MistralConfig(
        num_layers=2,
        seq_len=128,
        hidden_dim=16,
        num_heads=4,
        num_kv_heads=num_kv_heads,
        gradient_checkpointing=False,  # disable for tests so debugging is easier
        use_flash_attention=use_flash,
        flash_attention_block_size=8 if use_flash else None,
    )


@parameterize_with_configs("mistral*.yaml")
def test_mistral_configs(config_file):
    from levanter.main.train_lm import TrainLmConfig

    config_class = TrainLmConfig

    check_load_config(config_class, config_file)


@pytest.mark.parametrize("num_kv_heads", [1, 2])
def test_pass_different_length_seq(num_kv_heads):
    config = MistralConfig(
        seq_len=64,
        hidden_dim=32,
        intermediate_dim=32,
        num_heads=2,
        num_kv_heads=num_kv_heads,
        use_flash_attention=True,
    )
    check_model_works_with_seqlen(MistralLMHeadModel, config, 16)



import asyncio
import warnings
from typing import Mapping, Optional, Sequence, TypeVar

import jax
import numpy as np
from async_lru import alru_cache
from jax.random import PRNGKey
from jaxtyping import PRNGKeyArray

from haliax.util import StringHolderEnum

from levanter.data import AsyncDataset
from levanter.utils.index import Index
from levanter.utils.thread_utils import future_from_value


T = TypeVar("T")


class StopStrategy(metaclass=StringHolderEnum):
    FIRST_STOP_STRATEGY = "first_exhausted"
    ALL_STOP_STRATEGY = "all_exhausted"
    RESTART_STRATEGY = "restart"


class MixtureDataset(AsyncDataset[T]):
    """
    MixtureDataset supports loading data from multiple datasets. It takes a list of datasets and yields from them
    according to the weights.

    Creating a random-access MixtureDataset is challenging because we need to keep track of the current index of each
    dataset. So solve this, we instead use "block-deterministic" mixtures, where the number of samples from each dataset
    in each block is always identical (and we shuffle the order of the dataset ids in each block).

    Args:
        datasets: A dict of datasets, where the key is the name of the dataset and the value is the dataset itself
        weights: weights for each dataset
        stop_strategy: strategy for stopping the iteration, by default RESTART_STRATEGY. (Currently only RESTART_STRATEGY is supported)
            - FIRST_STOP_STRATEGY: stop when one dataset has been exhausted
            - ALL_STOP_STRATEGY: stop when all datasets have been exhausted
            - RESTART_STRATEGY: restart the dataset when it has been exhausted
        key: random key for datasets sampling
    """

    def __init__(
        self,
        datasets: Mapping[str, AsyncDataset[T]],
        weights: dict[str, float],
        block_size: int,
        *,
        randomize_blocks: bool = True,
        key: PRNGKeyArray | int,
        stop_strategy: str = StopStrategy.RESTART_STRATEGY,
    ):
        super().__init__()
        self.weights = MixtureDataset._normalize_weights(weights)
        self.datasets = {name: dataset for name, dataset in datasets.items() if self.weights.get(name, 0) > 0}
        self.dataset_index = Index(self.datasets.keys())
        self.block_size = block_size
        # we pack index and ds id into a single 32 bit, so block size must be at most 2^16
        if block_size >= 2**16:
            raise ValueError(f"Block size must be at most 2^16, got {block_size}")

        self.randomize_blocks = randomize_blocks

        if isinstance(key, int):
            key = PRNGKey(key)

        self.key = key

        if stop_strategy not in StopStrategy:  # type: ignore
            raise ValueError(f"Stop strategy {stop_strategy} is not supported.")

        # for now, just support restart strategy
        if stop_strategy != StopStrategy.RESTART_STRATEGY:
            raise NotImplementedError("Only restart strategy is supported for now.")

        self.stop_strategy = stop_strategy

        self._counts_per_block = self._compute_expected_counts_per_block(block_size)
        # precompute a list of ids for each block
        # the ids contain both the dataset index and the index within the dataset
        self._unpermuted_ids = self._compute_unpermuted_ids(self._counts_per_block)

    def _compute_expected_counts_per_block(self, block_size):
        _expected_values_per_block = np.zeros(len(self.datasets), dtype=np.int32)
        for i, dsname in enumerate(self.dataset_index):
            _expected_values_per_block[i] = self.weights[dsname] * block_size

        # handle remainder by adding to the largest dataset
        largest_dataset = np.argmax(_expected_values_per_block)
        _expected_values_per_block[largest_dataset] += block_size - _expected_values_per_block.sum()

        # check if any dataset has 0 samples (and nonzero weight)
        for i, dsname in enumerate(self.dataset_index):
            if _expected_values_per_block[i] == 0 and self.weights[dsname] > 0:
                warnings.warn(
                    f"Dataset {dsname} has 0 samples in the block, but weight of {self.weights[dsname]}."
                    " Recommend increasing block size."
                )

        return _expected_values_per_block

    def _compute_unpermuted_ids(self, counts_per_block):
        unpermuted_ids = np.zeros(int(counts_per_block.sum()), dtype=np.int64)
        start = 0
        for i, dsname in enumerate(self.dataset_index):
            count = counts_per_block[i]
            unpermuted_ids[start : start + count] = (i << 16) + np.arange(count)
            start += count
        return unpermuted_ids

    @staticmethod
    def _normalize_weights(weights: dict[str, float]):
        """Normalize the weights to sum to 1"""
        total = sum(weights.values())
        if total == 0:
            raise ValueError(f"Datasets' weights cannot sum to 0, got {weights}")
        return {name: weight / total for name, weight in weights.items() if weight > 0}

    async def async_len(self) -> int:
        if self.stop_strategy == StopStrategy.RESTART_STRATEGY:
            raise ValueError("Length is infinite for restart strategy")

        raise NotImplementedError("Length is not implemented for other strategies")

    async def final_length_is_known(self) -> bool:
        if self.stop_strategy == StopStrategy.RESTART_STRATEGY:
            return False

        raise NotImplementedError("Length is not known for other strategies")

    def is_finite(self) -> bool:
        if self.stop_strategy == StopStrategy.RESTART_STRATEGY:
            return False

        return True

    async def current_len(self) -> Optional[int]:
        if self.stop_strategy == StopStrategy.RESTART_STRATEGY:
            return None

        raise NotImplementedError("Length is not known for other strategies")

    @alru_cache
    async def _get_block(self, index: int) -> Optional[np.ndarray]:
        if not self.randomize_blocks:
            return self._unpermuted_ids

        return np.array(_compute_block_assignment(self._unpermuted_ids, index, self.key))

    def _index_into_dataset_for_id(self, id: int, block_id) -> tuple[int, int]:
        dataset_id = id >> 16
        dataset_index = id & 0xFFFF
        return dataset_id, dataset_index + block_id * self._counts_per_block[dataset_id]

    async def get_batch(self, indices: Sequence[int]) -> Sequence[T]:
        block_ids = np.array([idx // self.block_size for idx in indices])
        blocks = [self._get_block(block_id) for block_id in block_ids]
        blocks = await asyncio.gather(*blocks)

        # split the indices into batches for each dataset
        batches_per_dataset: list[list[int]] = [[] for _ in range(len(self.datasets))]
        indices_in_final_batch: list[list[int]] = [[] for _ in range(len(self.datasets))]

        assert len(indices) == len(blocks) == len(block_ids)

        for batch_index, (idx, block, block_id) in enumerate(zip(indices, blocks, block_ids)):
            index_within_block = idx % self.block_size  # which element of the block to get
            id = block[index_within_block]  # for this block, which dataset+base dataset offset
            dataset_id, dataset_index = self._index_into_dataset_for_id(id, block_id)
            batches_per_dataset[dataset_id].append(dataset_index)
            indices_in_final_batch[dataset_id].append(batch_index)

        # get the batches from each dataset
        batch_futures = []
        for dataset_id, indices_for_dataset in enumerate(batches_per_dataset):
            if len(indices_for_dataset) == 0:
                batch_futures.append(future_from_value([]))
            else:
                dataset = self._dataset_of_id(dataset_id)
                indices_for_dataset = await self._remap_indices(dataset, indices_for_dataset)
                batch_futures.append(dataset.get_batch(indices_for_dataset))

        batches = await asyncio.gather(*batch_futures)

        # reassemble the final batch
        final_batch = [None] * len(indices)

        for dataset_id, indices_into_batch in enumerate(indices_in_final_batch):
            for i, idx in enumerate(indices_into_batch):
                assert final_batch[idx] is None
                assert len(final_batch) > idx
                final_batch[idx] = batches[dataset_id][i]

        return final_batch  # type: ignore

    async def getitem_async(self, index: int) -> T:
        # simpler implementation because there's only one
        block_id = index // self.block_size
        index = index % self.block_size
        permuted_ids = await self._get_block(block_id)
        dataset_id, dataset_index = self._index_into_dataset_for_id(permuted_ids[index], block_id)

        dataset = self._dataset_of_id(dataset_id)
        dataset_index = (await self._remap_indices(dataset, [dataset_index]))[0]

        return await dataset.getitem_async(dataset_index)

    async def _remap_indices(self, ds, indices_into_ds):
        """
        Handles wrap around for datasets that have finite length
        """
        if self.stop_strategy == StopStrategy.RESTART_STRATEGY:
            if ds.is_finite():
                max_elem = max(indices_into_ds)
                length_of_dataset = await ds.wait_until_len_at_least(max_elem + 1)
                indices_into_ds = [idx % length_of_dataset for idx in indices_into_ds]

            return indices_into_ds

        raise NotImplementedError("Length is not known for other strategies")

    def _dataset_of_id(self, id):
        return self.datasets[self.dataset_index[id]]


def _compute_block_assignment(base_ids, index, key):
    rng = jax.random.fold_in(key, index)
    permuted_ids = jax.random.permutation(rng, base_ids)
    return permuted_ids



# References:
# * Orbax: https://github.com/google/orbax/blob/11d2934ecfff77e86b5e07d0fef02b67eff4511b/orbax/checkpoint/pytree_checkpoint_handler.py#L312
import asyncio
import functools
import logging
import os
from functools import partial
from typing import Callable, Optional

import equinox
import jax
import jax.experimental.array_serialization.serialization as array_ser
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
import tensorstore
from jax.sharding import Mesh
from tensorstore import TensorStore

import haliax as hax
import haliax.tree_util as htu
from haliax.partitioning import ResourceMapping
from haliax.util import is_named_array

from levanter.utils import jax_utils


logger = logging.getLogger(__name__)


def _is_named_or_none(x):
    return x is None or is_named_array(x)


def tree_serialize_leaves_tensorstore(
    checkpoint_dir,
    pytree,
    manager: Optional[array_ser.GlobalAsyncCheckpointManager] = None,
    *,
    commit_callback: Optional[Callable] = None,
):
    if manager is None:
        manager = array_ser.GlobalAsyncCheckpointManager()
        manager_was_none = True
    else:
        manager_was_none = False

    leaf_key_paths = jax_utils.leaf_key_paths(pytree, is_leaf=_is_named_or_none)

    def path_from_key_path(key_path):
        return os.path.join(checkpoint_dir, *key_path.split("."))

    paths = jtu.tree_map(path_from_key_path, leaf_key_paths, is_leaf=lambda x: x is None)
    paths = jtu.tree_leaves(paths, is_leaf=lambda x: x is None)
    leaves = jtu.tree_leaves(pytree, is_leaf=lambda x: x is None)
    assert len(leaves) == len(paths)

    # ok, not all of these are arrays, but we'll deal with that in the async function
    def _ensure_is_array(x):
        if isinstance(x, (int, float, bool, complex)):
            return jnp.array(x)
        else:
            return x

    arrays = [_ensure_is_array(x) for x in leaves]

    # filter out the None leaves and paths (must be zip)
    arrays, paths = zip(*[(a, p) for a, p in zip(arrays, paths) if equinox.is_array_like(a)])

    arrays = list(arrays)
    paths = list(paths)

    if commit_callback is None:
        commit_callback = lambda: logger.info("Committed checkpoint to Tensorstore")  # noqa

    manager.serialize_with_paths(arrays, paths, on_commit_callback=commit_callback)

    if manager_was_none:
        manager.wait_until_finished()


def _tensorstore_spec_for(checkpoint_dir, key_path: str):
    checkpoint_path = os.path.join(checkpoint_dir, *key_path.split("."))
    ts_spec = array_ser.get_tensorstore_spec(checkpoint_path)
    return ts_spec


async def _serialize_one_leaf(x, spec):
    if isinstance(x, hax.NamedArray):
        # we don't need to do anything special for named arrays to serialize, though we will for deserialization.
        return await _serialize_one_leaf(x.array, spec)
    elif isinstance(x, jax.Array):
        if not x.is_fully_addressable:
            return await array_ser.async_serialize(x, spec)
        else:
            return await save_array_to_tensorstore(x, spec)
    elif isinstance(x, (bool, float, complex, int)):
        return await save_array_to_tensorstore(np.array(x), spec)
    elif x is None:
        return
    elif isinstance(x, jnp.ndarray):
        return await save_array_to_tensorstore(x, spec)
    elif isinstance(x, np.ndarray):
        return await save_array_to_tensorstore(x, spec)
    else:
        raise TypeError(f"Can't serialize {type(x)}")


async def save_array_to_tensorstore(x, spec):
    if jax.process_index() == 0:
        if x.dtype == jnp.bfloat16:
            # Tensorstore uses 'bfloat16', not '<V2'.
            dtype = "bfloat16"
        else:
            dtype = np.dtype(x.dtype).str
        t = await tensorstore.open(
            tensorstore.Spec(spec), create=True, shape=x.shape, dtype=dtype, context=array_ser.TS_CONTEXT
        )

        await t.write(x)


async def load_array_from_tensorstore(spec):
    t: TensorStore = await tensorstore.open(tensorstore.Spec(spec), context=array_ser.TS_CONTEXT)
    return await t.read()


async def _deserialize_one_leaf(like, spec, axis_mapping, mesh):
    if is_named_array(like):
        return await _deserialize_named_array(like, spec, axis_mapping, mesh)
    elif isinstance(like, jax.Array):
        if not like.is_fully_addressable:
            return await array_ser.async_deserialize(like.sharding, spec, global_shape=like.shape, dtype=like.dtype)
        else:
            return await load_array_from_tensorstore(spec)
    elif isinstance(like, (bool, float, complex, int)):
        arr = await load_array_from_tensorstore(spec)
        return arr.item()
    elif like is None:
        return None
    elif isinstance(like, jnp.ndarray) or isinstance(like, np.ndarray) or isinstance(like, jax.ShapeDtypeStruct):
        return await load_array_from_tensorstore(spec)
    elif callable(like):
        return like
    else:
        raise TypeError(f"Can't deserialize {type(like)}")


async def _deserialize_named_array(like, spec, axis_mapping, mesh):
    # the main thing we're worried about is deserialized NamedArrays that are not yet arrays but are ShapedDtypeStructs.
    # These don't (currently) have sharding info, but we can infer it from the axes
    if isinstance(like.array, jax.ShapeDtypeStruct):
        sharding = hax.partitioning.sharding_for_axis(like.axes, axis_mapping, mesh)
        array = await array_ser.async_deserialize(sharding, spec, global_shape=like.array.shape, dtype=like.dtype)
        assert sharding.is_equivalent_to(array.sharding, len(like.array.shape))
        return hax.NamedArray(array, like.axes)
    else:
        array = await _deserialize_one_leaf(like.array, spec, axis_mapping, mesh)
        return hax.NamedArray(array, like.axes)


def tree_deserialize_leaves_tensorstore(
    checkpoint_dir, pytree, axis_mapping: Optional[ResourceMapping] = None, mesh: Optional[Mesh] = None
):
    """
    Deserializes a PyTree of Arrays and NamedArrays from a Tensorstore checkpoint, returning a pytree with the same shape
    as the one provided. This method is capable of deserializing NamedArrays that are the result of an eval_shape call
    (i.e. they are not yet arrays but are ShapedDtypeStructs), provided you pass in the axis_mapping and mesh (or
    they are available by context)

    :param checkpoint_dir: the directory containing the tensorstore checkpoint, can be a local path or a GCS path
    :param pytree: the exemplar pytree
    :param axis_mapping: optional, the axis mapping for the NamedArrays (if they are not yet arrays)
    :param mesh: optional, the mesh for the NamedArrays (if they are not yet arrays)

    :return: a pytree with the same shape as the exemplar pytree, but with the arrays deserialized from the checkpoint
    """
    # TODO: support ShapeDtypeStructs that are not NamedArrays
    leaf_key_paths = jax_utils.leaf_key_paths(pytree, is_leaf=is_named_array)
    specs = htu.tree_map(partial(_tensorstore_spec_for, checkpoint_dir), leaf_key_paths)

    deser_partial = functools.partial(_deserialize_one_leaf, axis_mapping=axis_mapping, mesh=mesh)

    futures = jtu.tree_map(deser_partial, pytree, specs, is_leaf=is_named_array)
    leaves, structure = jtu.tree_flatten(futures, is_leaf=is_named_array)

    async def _do_deserialize():
        values = await asyncio.gather(*leaves)
        return jtu.tree_unflatten(structure, values)

    return asyncio.run(_do_deserialize())



import logging
from dataclasses import dataclass, field

import equinox as eqx
import jax
import jmp

import haliax as hax
from haliax import Axis
from haliax.partitioning import fsdp, round_axis_for_partitioning

import levanter
from levanter.checkpoint import load_checkpoint
from levanter.data import DataLoader
from levanter.data.text import CausalLmDataset, LMDatasetConfig
from levanter.models.gpt2 import Gpt2Config
from levanter.models.lm_model import LmConfig, LmExample, LmHeadModel, compute_next_token_loss
from levanter.trainer import TrainerConfig
from levanter.utils.jax_utils import use_cpu_device
from levanter.utils.tree_utils import inference_mode
from levanter.visualization import compute_and_visualize_log_probs


logger = logging.getLogger(__name__)


@dataclass
class VizGpt2Config:
    checkpoint_path: str
    path: str = "logprobs.html"
    trainer: TrainerConfig = field(default_factory=TrainerConfig)
    data: LMDatasetConfig = field(default_factory=LMDatasetConfig)
    model: LmConfig = field(default_factory=Gpt2Config)

    num_docs: int = 256


def main(config: VizGpt2Config):
    levanter.initialize(config)
    tokenizer = config.data.the_tokenizer

    # some axes we use outside the model proper
    EvalBatch = config.trainer.EvalBatch
    Pos = config.model.Pos
    KeyPos = config.model.KeyPos

    eval_loader = DataLoader(
        EvalBatch,
        CausalLmDataset(config.data.validation_set(Pos.size), Pos, KeyPos),  # type: ignore
        32,
        config.trainer.device_mesh,
        config.trainer.compute_axis_mapping,
    )

    # some axes we use outside the model proper
    Pos = config.model.Pos

    compute_axis_mapping = config.trainer.compute_axis_mapping
    parameter_axis_mapping = config.trainer.parameter_axis_mapping

    with config.trainer.device_mesh, hax.axis_mapping(parameter_axis_mapping):
        key = jax.random.PRNGKey(0)

        vocab_size = len(tokenizer)
        Vocab = round_axis_for_partitioning(Axis("vocab", vocab_size), compute_axis_mapping)
        if vocab_size != Vocab.size:
            logger.info(f"Rounding vocab size from {vocab_size} to {Vocab.size} for partitioning")

        mp: jmp.Policy = config.trainer.mp

        # don't want to compute the mask w.r.t. the final token

        @fsdp(parameter_axis_mapping, compute_axis_mapping)
        def compute_log_probs(model: LmHeadModel, example: LmExample):
            model = inference_mode(model, True)
            model = mp.cast_to_compute(model)
            logprobs = compute_next_token_loss(model, example, reduction=None)
            # roll forward to get the loss for each predicted token
            logprobs = hax.roll(logprobs, 1, Pos)
            return logprobs.rearrange((EvalBatch, Pos)).array

        # initialize the model
        with use_cpu_device():
            model = eqx.filter_eval_shape(config.model.build, Vocab, key=key)
            # TODO: don't load the entire checkpoint into CPU memory when we only need our share of the model
            model = load_checkpoint(model, config.checkpoint_path, subpath="model")

        assert model is not None

        model = hax.shard_with_axis_mapping(model, parameter_axis_mapping)

        compute_and_visualize_log_probs(
            path=config.path,
            model=model,
            tokenizer=tokenizer,
            log_prob_fn=compute_log_probs,
            test_data=eval_loader,
            max_docs=config.num_docs,
        )


if __name__ == "__main__":
    levanter.config.main(main)()



import abc
import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Generic, Optional, Sequence, TypeVar

import jax.random
import numpy as np
from jax.random import PRNGKey

from levanter.utils import thread_utils


logger = logging.getLogger(__name__)


T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")
U = TypeVar("U")


_executor = ThreadPoolExecutor(max_workers=10)


class DatasetBase(abc.ABC, Generic[T_co]):
    """
    Base class for sync and async datasets. This class is not meant to be used directly.
    """

    @abc.abstractmethod
    def as_async_dataset(self) -> "AsyncDataset[T_co]":
        raise NotImplementedError("...")

    @abc.abstractmethod
    def as_sync_dataset(self) -> "SyncDataset[T_co]":
        raise NotImplementedError("...")


class AsyncDataset(DatasetBase[T_co]):
    """
    An asynchronous dataset that can be used with async/await syntax. In Levanter, we use AsyncDataset for two purposes:
    * To represent datasets that are inherently asynchronous (e.g. reading from disk, network, etc.).
    * To represent datasets that are still being constructed.

    The core methods in this class are:
    * `async_len`: Returns the final length of the dataset.
    * `get_batch`: Returns a batch of items from the dataset.
    * `current_len`: Returns the current length of the dataset. This may be None if no current length is known.
    """

    def __init__(self):
        self._min_known_len = 0

    @abc.abstractmethod
    async def async_len(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    async def final_length_is_known(self) -> bool:
        """Returns whether the final length of the dataset is known.
        If this returns False, the current_len of the dataset may change in the future."""
        raise NotImplementedError

    @abc.abstractmethod
    def is_finite(self) -> bool:
        """
        Returns whether the dataset will have a known length in the future (e.g. if it's being constructed).
        If this returns False, the length of the dataset is infinite or unknowable.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def current_len(self) -> Optional[int]:
        """
        Returns the current length of the dataset that won't require (expensive) waiting.

        If the current length is not known, returns None. This might block temporarily for a short time to get the
        current length.
        """
        raise NotImplementedError

    async def getitem_async(self, index: int) -> T_co:
        """
        Returns the item at the given index. Typically implemented as a wrapper around `get_batch`.

        In general, it is better to call (and override) `get_batch` instead of this method.
        """
        return (await self.get_batch([index]))[0]

    @abc.abstractmethod
    async def get_batch(self, indices: Sequence[int]) -> Sequence[T_co]:
        raise NotImplementedError

    async def wait_until_len_at_least(self, length: int) -> int:
        """
        Returns the length of the dataset once it is at least `length` or if the dataset has a known (finished) length.

        The default implementation is a naive busy-wait loop. You should override this method for more efficient
        implementations.
        """
        if length <= self._min_known_len:
            return self._min_known_len

        res_len = await naive_busy_wait_until_len_at_least(self, length)
        self._min_known_len = max(self._min_known_len, res_len)
        return res_len

    def as_sync_dataset(self):
        return SyncifiedDataset(self)

    def as_async_dataset(self) -> "AsyncDataset[T_co]":
        return self

    def map(self, fn: Callable[[T_co], U], *extra_args, **extra_kwargs) -> "MappedAsyncDataset[T_co, U]":
        return MappedAsyncDataset(self, fn, *extra_args, **extra_kwargs)

    def shuffle(self, key: PRNGKey):
        import levanter.data.permutation as permutation

        return permutation.PermutationDataset(self, key)

    def era_shuffle(self, era_length: int, key: PRNGKey):
        import levanter.data.permutation as permutation

        return permutation.EraShufflingDataset(self, era_length, key=key)


async def naive_busy_wait_until_len_at_least(dataset: AsyncDataset[T_co], length: int) -> int:
    """
    Runs a busy-wait loop until the dataset has at least `length` items or the final length is known.

    Returns the current length of the dataset when either the dataset has at least `length` items or the final length is
    known.

    You should probably implement this in a more efficient way. This is just a naive implementation.
    """
    while not await dataset.final_length_is_known():
        current_len = await dataset.current_len()
        if current_len is None:
            raise ValueError("Dataset has unknown length")
        if current_len <= length:
            await asyncio.sleep(0.1)
        else:
            return current_len

    return await dataset.async_len()


class SyncDataset(DatasetBase[T_co]):
    """
    A synchronous dataset that can be used with regular Python syntax. In Levanter, we mainly do not use this class.
    You can use this class if it's easier, then convert it to an AsyncDataset using `as_async_dataset`. This
    is not as efficient as using an AsyncDataset directly, but it can be useful for testing or for simpler code.
    """

    @abc.abstractmethod
    def __len__(self) -> int:
        """
        Returns the final length of the data store.
        May raise if the length is not known.
        """

    @abc.abstractmethod
    def has_len(self) -> bool:
        """
        Whether the data store currently has a known length. If this returns False, then the length of the data store
        may change in the future.
        """
        pass

    @abc.abstractmethod
    def current_len(self) -> Optional[int]:
        """
        Returns the current length of the data store. If the length is infinite or not known, returns None.
        """
        pass

    def __getitem__(self, index: int) -> T_co:
        return self.get_batch([index])[0]

    @abc.abstractmethod
    def get_batch(self, indices: Sequence[int] | np.ndarray) -> Sequence[T_co]:
        pass

    def as_async_dataset(self) -> "AsyncDataset[T_co]":
        return AsyncifiedDataset(self)

    def as_sync_dataset(self) -> "SyncDataset[T_co]":
        return self


class SyncifiedDataset(SyncDataset[T_co]):
    def __init__(self, dataset: AsyncDataset[T_co]):
        self.dataset = dataset

    def _run_coroutine(self, coro):
        return thread_utils.blocking_wait(coro)

    def __len__(self) -> int:
        return self._run_coroutine(self.dataset.async_len())

    def has_len(self) -> bool:
        return self.dataset.is_finite()

    def current_len(self) -> Optional[int]:
        return self._run_coroutine(self.dataset.current_len())

    def get_batch(self, indices: Sequence[int] | np.ndarray) -> Sequence[T_co]:
        return self._run_coroutine(self.dataset.get_batch(indices))

    def __getitem__(self, index: int) -> T_co:
        return self._run_coroutine(self.dataset.getitem_async(index))


class AsyncifiedDataset(AsyncDataset[T_co]):
    def __init__(self, dataset: SyncDataset[T_co]):
        super().__init__()
        self.dataset = dataset

    async def async_len(self) -> int:
        return len(self.dataset)

    async def final_length_is_known(self) -> bool:
        return self.dataset.has_len()

    def is_finite(self) -> bool:
        return self.dataset.has_len()

    async def current_len(self) -> Optional[int]:
        return self.dataset.current_len()

    async def get_batch(self, indices: Sequence[int]) -> Sequence[T_co]:
        return self.dataset.get_batch(indices)

    async def getitem_async(self, index: int) -> T_co:
        return self.dataset[index]

    def __repr__(self):
        return f"WrappedAsyncDataset({repr(self.dataset)})"

    def __str__(self):
        return f"WrappedAsyncDataset({str(self.dataset)})"


class ListAsyncDataset(AsyncDataset[T]):
    """
    A simple dataset that wraps a list. Mostly for testing.
    """

    def __init__(self, data: list[T], is_complete: bool = False):
        super().__init__()
        self.data = data
        self.is_complete = is_complete
        if not is_complete:
            self.complete_promise: Optional[asyncio.Future[None]] = asyncio.Future()
            self.length_updated: Optional[asyncio.Condition] = asyncio.Condition()
        else:
            self.complete_promise = None
            self.length_updated = None

    async def async_len(self) -> int:
        # this is the final length
        if not self.is_complete:
            assert self.complete_promise is not None
            await self.complete_promise
        return len(self.data)

    async def final_length_is_known(self) -> bool:
        return self.is_complete

    def is_finite(self) -> bool:
        return True

    async def current_len(self) -> Optional[int]:
        return len(self.data)

    async def get_batch(self, indices: Sequence[int]) -> Sequence[T]:
        await self.wait_until_len_at_least(max(indices) + 1)
        return [self.data[i] for i in indices]

    def append(self, item: T):
        if self.is_complete:
            raise ValueError("Cannot append to a finalized dataset")
        self.data.append(item)
        asyncio.create_task(self.notify_length_update())

    def finalize(self):
        self.is_complete = True
        if self.complete_promise is not None:
            self.complete_promise.set_result(None)
            if not asyncio.get_event_loop().is_running():
                _executor.submit(lambda: asyncio.run(self.notify_length_update()))
            else:
                asyncio.create_task(self.notify_length_update())

    async def notify_length_update(self):
        async with self.length_updated:
            self.length_updated.notify_all()

    async def wait_until_len_at_least(self, length: int) -> int:
        if self.is_complete:
            return len(self.data)

        assert self.length_updated is not None

        async with self.length_updated:
            while len(self.data) < length and not self.is_complete:
                await self.length_updated.wait()

        return len(self.data)


class MappedAsyncDataset(AsyncDataset[U], Generic[T, U]):
    """
    A dataset that applies a function to each item in the dataset.
    You can pass extra arguments to the function using `*extra_args` and `**extra_kwargs`.
    If a kwarg called `key` is passed, it will be treated as a PRNGKey and folded in with the index of the item
    for each call to the function.
    """

    def __init__(
        self,
        dataset: AsyncDataset[T],
        fn: Callable[[T], U] | Callable[[T, Optional[PRNGKey]], U],
        *extra_args,
        **extra_kwargs,
    ):
        super().__init__()
        self.dataset = dataset
        self.fn = fn
        self._extra_args = extra_args
        self._extra_kwargs = extra_kwargs

    async def async_len(self) -> int:
        return await self.dataset.async_len()

    async def final_length_is_known(self) -> bool:
        return await self.dataset.final_length_is_known()

    def is_finite(self) -> bool:
        return self.dataset.is_finite()

    async def current_len(self) -> Optional[int]:
        return await self.dataset.current_len()

    def _maybe_fold_in_key(self, key, index):
        if key is not None:
            key = jax.random.fold_in(key, index)
        return key

    async def get_batch(self, indices: Sequence[int]) -> Sequence[U]:
        items = await self.dataset.get_batch(indices)
        return [self._call_fn(i, item) for i, item in zip(indices, items)]

    async def getitem_async(self, index: int) -> U:
        return self._call_fn(index, await self.dataset.getitem_async(index))

    async def wait_until_len_at_least(self, length: int) -> int:
        return await self.dataset.wait_until_len_at_least(length)

    def _call_fn(self, index, item):
        if "key" in self._extra_kwargs:
            key = self._maybe_fold_in_key(self._extra_kwargs["key"], index)
            kwargs = {**self._extra_kwargs, "key": key}
        else:
            kwargs = self._extra_kwargs
        return self.fn(item, *self._extra_args, **kwargs)



import dataclasses
import datetime
import pathlib
import tempfile
from datetime import timedelta

import equinox as eqx
import jax
import jax.tree_util as jtu
import numpy as np
import optax
from chex import assert_trees_all_close, assert_trees_all_equal
from jax import ShapeDtypeStruct
from jax import numpy as jnp

import haliax
from haliax import Axis

from levanter.checkpoint import (
    Checkpointer,
    CheckpointInterval,
    discover_latest_checkpoint,
    load_checkpoint,
    load_checkpoint_or_initialize,
    load_metadata,
    save_checkpoint,
)
from levanter.trainer import StepInfo
from levanter.trainer_state import TrainerState
from test_utils import MLP, arrays_only, assert_trees_not_close


def _dummy_step_info(step):
    return StepInfo(
        state=TrainerState(
            # + 1 b/c step here is next step
            step=step + 1,
            model=None,
            optimizer=None,  # type: ignore
            opt_state=None,
            training_key=jax.random.PRNGKey(0),
            is_trainable=True,
            mp=None,
        ),
        loss=0.0,
        step_duration=0.0,
    )


def _get_checkpoint_steps(checkpoint_dir):
    paths = list(pathlib.Path(checkpoint_dir).iterdir())
    return sorted([load_metadata(f)["step"] for f in paths])


def test_checkpointer_changing_policy():
    with tempfile.TemporaryDirectory(prefix="checkpoints") as tmpdir:
        checkpointer = Checkpointer(
            tmpdir,
            None,
            [
                CheckpointInterval(every=2, until=10),
                CheckpointInterval(every=5, until=20),
                CheckpointInterval(every=10, until=None),
            ],
        )

        for step in range(1, 50):
            checkpointer.on_step(_dummy_step_info(step))

        checkpointer.wait_until_finished()

        # ensure we saved the right checkpoints
        assert _get_checkpoint_steps(tmpdir) == [2, 4, 6, 8, 10, 15, 20, 30, 40]


def test_checkpointer_temporal_policy():
    fake_now = datetime.datetime(2021, 1, 1, 0, 0, 0)

    tick = 10

    def advance_time(delta_seconds):
        nonlocal fake_now
        fake_now += timedelta(seconds=delta_seconds)

    with tempfile.TemporaryDirectory(prefix="checkpoints") as tmpdir:
        checkpointer = Checkpointer(tmpdir, timedelta(seconds=tick), [], dt_now_injection=lambda: fake_now)

        checkpointer.on_step(_dummy_step_info(0))
        advance_time(tick)
        checkpointer.on_step(_dummy_step_info(1))
        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [1]

        advance_time(tick - 1)
        checkpointer.on_step(_dummy_step_info(2))
        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [1]
        advance_time(1)
        checkpointer.on_step(_dummy_step_info(3))
        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [3]


def test_checkpointer_mixed_policy():
    fake_now = datetime.datetime(2021, 1, 1, 0, 0, 0)

    tick = 10

    def advance_time(delta_seconds):
        nonlocal fake_now
        fake_now += timedelta(seconds=delta_seconds)

    with tempfile.TemporaryDirectory(prefix="checkpoints") as tmpdir:
        checkpointer = Checkpointer(
            tmpdir,
            timedelta(seconds=tick),
            [
                CheckpointInterval(every=2, until=10),
                CheckpointInterval(every=5, until=20),
                CheckpointInterval(every=10, until=None),
            ],
            dt_now_injection=lambda: fake_now,
        )

        checkpointer.on_step(_dummy_step_info(0))
        advance_time(tick)
        checkpointer.on_step(_dummy_step_info(1))
        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [1]

        advance_time(tick - 1)
        # time hasn't advanced enough, so we wouldn't save a checkpoint, but we do because of the interval
        checkpointer.on_step(_dummy_step_info(2))
        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [2]

        advance_time(1)
        # time has advanced enough now from last temporal save, but we don't save a checkpoint because we just saved one
        checkpointer.on_step(_dummy_step_info(3))
        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [2]

        for step in range(4, 11):
            advance_time(tick)
            checkpointer.on_step(_dummy_step_info(step))
            # we need this to stop a race condition

        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [2, 4, 6, 8, 10]

        advance_time(tick - 1)
        checkpointer.on_step(_dummy_step_info(11))
        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [2, 4, 6, 8, 10]

        for step in range(12, 50):
            checkpointer.on_step(_dummy_step_info(step))
            advance_time(tick)

        # ensure we saved the right checkpoints
        checkpointer.wait_until_finished()
        assert _get_checkpoint_steps(tmpdir) == [2, 4, 6, 8, 10, 15, 20, 30, 40, 49]  # 49 is last temporary checkpoint


def _make_state(step, key):
    model = MLP(in_size=2, out_size=1, width_size=2, depth=3, key=key)
    optim = optax.adam(1e-4)
    opt_state = optim.init(arrays_only(model))

    return TrainerState(step, model, optim, opt_state, key, is_trainable=True, mp=None)


def test_checkpoint_simple():
    key0 = jax.random.PRNGKey(0)
    key1 = jax.random.PRNGKey(1)

    initial_state = _make_state(10, key0)
    rep_state = _make_state(2, key1)

    assert_trees_not_close(initial_state.model, rep_state.model)

    with tempfile.TemporaryDirectory() as tmpdir:
        save_checkpoint(
            initial_state,
            step=initial_state.step,
            checkpoint_path=tmpdir,
        )
        restored_state = load_checkpoint(
            rep_state,
            checkpoint_path=tmpdir,
            discover_latest=False,
        )

        assert_trees_all_equal(
            jax.tree_util.tree_leaves(arrays_only(restored_state.model)),
            jax.tree_util.tree_leaves(arrays_only(initial_state.model)),
        )
        assert all(np.isclose(restored_state.training_key, initial_state.training_key))
        assert restored_state.step == initial_state.step


def test_checkpoint_steps():
    key0 = jax.random.PRNGKey(0)
    key1 = jax.random.PRNGKey(1)

    optim = optax.adam(1e-4)

    initial_state = _make_state(10, key0)
    data = jax.random.uniform(key0, (2, 2))

    @eqx.filter_grad
    def loss_fn(model, data):
        m = jax.vmap(model)
        return jnp.mean(jnp.square(m(data)))

    state = initial_state
    for i in range(3):
        grad = loss_fn(state.model, data)
        updates, new_state = optim.update(grad, state.opt_state)
        model = eqx.apply_updates(state.model, updates)
        state = dataclasses.replace(state, step=state.step + 1, model=model, opt_state=new_state)

    assert_trees_not_close(state, initial_state)

    rep_state = _make_state(42, key1)
    assert_trees_not_close(state, rep_state)

    with tempfile.TemporaryDirectory() as tmpdir:
        save_checkpoint(state, step=3, checkpoint_path=tmpdir)
        restored_state = load_checkpoint(rep_state, checkpoint_path=tmpdir, discover_latest=False)

        assert_trees_all_equal(
            jax.tree_util.tree_leaves(arrays_only(restored_state)),
            jax.tree_util.tree_leaves(arrays_only(state)),
        )


def test_checkpoint_discovery():
    with tempfile.TemporaryDirectory() as tempdir:
        save_checkpoint(dict(model=1, training_state=2), step=10, checkpoint_path=f"{tempdir}/step-10")
        save_checkpoint(dict(model=3, training_state=4), step=20, checkpoint_path=f"{tempdir}/step-20")
        save_checkpoint(dict(model=5, training_state=6), step=30, checkpoint_path=f"{tempdir}/step-30")

        latest = discover_latest_checkpoint(tempdir)
        assert latest == f"{tempdir}/step-30"

        assert discover_latest_checkpoint("file:///tmp/does-not-exist") is None


def test_load_from_checkpoint_or_initialize():
    In = Axis("in", 2)
    Out = Axis("out", 1)

    def init_fn(key):
        return haliax.nn.MLP.init(In, Out, 2, 1, key=key, use_bias=False, use_final_bias=False)

    k0 = jax.random.PRNGKey(0)
    k1 = jax.random.PRNGKey(1)

    model0 = eqx.filter_jit(init_fn)(k0)
    model1 = eqx.filter_jit(init_fn)(k1)

    is_checkpointed = jtu.tree_map(lambda _: False, model0)
    is_checkpointed = eqx.tree_at(lambda t: t.layers[-1], is_checkpointed, replace=True)
    is_checkpointed1 = jtu.tree_map(lambda _: False, model1)
    is_checkpointed1 = eqx.tree_at(lambda t: t.layers[-1], is_checkpointed1, replace=True)

    with jax.sharding.Mesh(jax.devices(), ("devices",)), tempfile.TemporaryDirectory() as tmpdir:
        filtered = eqx.filter(model0, is_checkpointed)
        save_checkpoint(filtered, step=0, checkpoint_path=tmpdir)

        loaded = load_checkpoint_or_initialize(init_fn, tmpdir, is_checkpointed=is_checkpointed, donate_args=False)(k1)
        assert not any(jax.tree_util.tree_leaves(eqx.filter(loaded, lambda x: isinstance(x, ShapeDtypeStruct))))

        loaded2 = load_checkpoint(eqx.filter(model1, is_checkpointed), tmpdir, discover_latest=True)
        loaded2 = eqx.combine(loaded2, model1)

        assert_trees_all_equal(
            jax.tree_util.tree_leaves(arrays_only(loaded)),
            jax.tree_util.tree_leaves(arrays_only(loaded2)),
        )

        assert_trees_all_equal(
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(loaded, is_checkpointed))),
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(model0, is_checkpointed))),
        )

        assert_trees_not_close(
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(loaded, is_checkpointed, inverse=True))),
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(model0, is_checkpointed, inverse=True))),
        )

        assert_trees_not_close(
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(loaded, is_checkpointed))),
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(model1, is_checkpointed))),
        )

        assert_trees_all_equal(
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(loaded, is_checkpointed, inverse=True))),
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(model1, is_checkpointed1, inverse=True))),
        )


def test_load_from_checkpoint_or_initialize_works_if_file_not_found():
    In = Axis("in", 2)
    Out = Axis("out", 1)

    def init_fn(key):
        return haliax.nn.MLP.init(In, Out, 2, 3, key=key)

    k0 = jax.random.PRNGKey(0)
    k1 = jax.random.PRNGKey(1)

    model0 = init_fn(k0)
    model1 = init_fn(k1)

    is_checkpointed = jtu.tree_map(lambda _: False, model0)
    is_checkpointed = eqx.tree_at(lambda t: t.layers[-1], is_checkpointed, replace=True)

    with jax.sharding.Mesh(jax.devices(), ("devices",)):
        loaded = load_checkpoint_or_initialize(init_fn, "kanmfklafnmjlkanfjklanfjkh", is_checkpointed=is_checkpointed)(
            k1
        )

        assert not any(jax.tree_util.tree_leaves(eqx.filter(loaded, lambda x: isinstance(x, ShapeDtypeStruct))))
        # should be the same as model1
        # on TPU, there's a very slight difference for some reason
        assert_trees_all_close(
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(loaded, is_checkpointed))),
            jax.tree_util.tree_leaves(arrays_only(eqx.filter(model1, is_checkpointed))),
        )



import contextlib
import json
import logging
import os
import shutil
import tempfile
import time
import urllib.parse
from typing import Optional

import fsspec
import jax
import requests  # type: ignore
from fsspec import AbstractFileSystem
from jax.experimental.multihost_utils import sync_global_devices


logger = logging.getLogger(__name__)

SENTINEL_FILE = "/tmp/tpu_shutdown_sentinel"


def _checked_request(url):
    try:
        response = requests.get(url, headers={"Metadata-Flavor": "Google"})
        response.raise_for_status()
        return response.text
    except requests.exceptions.RequestException:
        logger.exception(f"Could not get {url} from metadata server. Is this a TPU VM?", exc_info=True)
        raise


def _checked_delete(url):
    # first get the token
    token = _checked_request(
        "http://metadata.google.internal/computeMetadata/v1/instance/service-accounts/default/token"
    )
    token = json.loads(token)["access_token"]
    headers = {"Authorization": f"Bearer {token}", "Metadata-Flavor": "Google"}
    try:
        response = requests.delete(url, headers=headers)
        response.raise_for_status()
        return response.text
    except requests.exceptions.RequestException:
        logger.exception(f"Could not delete {url} from metadata server. Is this a TPU VM?", exc_info=True)
        raise


def _shutdown_tpu_with_queued_resource():
    queued_resource = _checked_request(
        "http://metadata.google.internal/computeMetadata/v1/instance/attributes/queued-resource-name"
    )
    # queued resource looks like:
    # projects/999999/locations/us-central2-b/queuedResources/NAME
    # to delete we need to use delete against
    # https://tpu.googleapis.com/v2/projects/9999/locations/us-central2-b/queuedResources/NAME?force=true
    if queued_resource:
        queued_resource_name = queued_resource.split("/")[-1]
        # quiet really works like -y
        if jax.process_index() == 0:
            logger.critical(f"Found queued resource {queued_resource_name}. Attempting to delete it.")
            # We need to use curl
            # curl -X DELETE -H "Authorization: Bearer $(gcloud auth print-access-token)" \
            # -H "Content-Type: application/json" \
            # https://tpu.googleapis.com/v2/projects/my-project/locations/us-central2-b/queuedResources/my-queued-resource?force=true
            # os.system(f"gcloud compute tpus queued-resources delete {queued_resource} --zone {zone} --force --quiet")
            url = f"https://tpu.googleapis.com/v2/{queued_resource}?force=true"
            _checked_delete(url)
        return True
    else:
        logger.info("No queued resource found.")
        return False


def shutdown_tpu_vm(sleep_seconds=60 * 5):
    """You should probably call this from atexit or something like that."""
    # fork a process to do the delete so the main process can exit before the delete is done
    logger.info("Forking a process to delete...")
    logger.critical(f"Create a file {SENTINEL_FILE} to cancel the shutdown")
    logger.critical(f"$ touch {SENTINEL_FILE}")

    # fork works better for our use case
    pid = os.fork()
    if pid == 0:
        _do_shutdown_tpu_vm(sleep_seconds)
        os._exit(0)
    else:
        logger.info(f"Forked process {pid} to delete TPU VM")


def _do_shutdown_tpu_vm(sleep_seconds):
    # the gcloud command we would run is something like:
    # gcloud compute tpus tpu-vm delete tpu-vm-1 --zone us-central1-a --quiet
    try:
        zone = _checked_request("http://metadata.google.internal/computeMetadata/v1/instance/zone")
        zone = zone.split("/")[-1]
        name = _checked_request("http://metadata.google.internal/computeMetadata/v1/instance/attributes/instance-id")
        project = _checked_request("http://metadata.google.internal/computeMetadata/v1/project/project-id")
    except requests.exceptions.RequestException:
        logger.warning("Could not get zone or instance-id from metadata server. Is this a TPU VM? Not shutting down.")
        return

    logger.critical(f"Shutting down TPU VM {name} in zone {zone} in {sleep_seconds} seconds")
    time.sleep(sleep_seconds)
    if os.path.exists(SENTINEL_FILE):
        logger.critical(f"Found sentinel file {SENTINEL_FILE}, not shutting down TPU VM")
        return
    logger.critical(f"Shutting down TPU VM {name} in zone {zone}")

    try:
        success = _shutdown_tpu_with_queued_resource()
        if success:
            return
    except requests.exceptions.RequestException:
        logger.info("This is not a queued resource, deleting the old fashioned way.")

    logger.critical(f"Shutting down TPU VM {name} in zone {zone}")
    if jax.process_index() != 0:
        logger.info(f"Letting process 0 handle the shutdown. We are process {jax.process_index()}")
        return

    # os.system(f"gcloud compute tpus tpu-vm delete {name} --zone {zone} --quiet")
    # https://tpu.googleapis.com/v2/projects/PROJECT/locations/us-central2-b/nodes/NAME
    url = f"http://tpu.googleapis.com/v2/projects/{project}/locations/{zone}/nodes/{name}"
    _checked_delete(url)


_sync_count = 0


@contextlib.contextmanager
def temp_dir_before_upload(path, process_should_upload: Optional[bool] = None):
    """
    Creates a temp_dir, yields it, then uploads it to the given path or url on exit using fsspec.
    If it's a local path instead of a url, it just yields the path.

    :param process_should_upload: If None, only upload if jax.process_index() == 0. Otherwise, upload if this is True.
    """

    if process_should_upload is None:
        process_should_upload = jax.process_index() == 0

    def _is_url_like(path):
        return urllib.parse.urlparse(path).scheme != ""

    try:
        tmpdir: Optional[str] = None
        if _is_url_like(path):
            tmpdir = tempfile.mkdtemp()
            local_path = tmpdir
        else:
            local_path = path

        yield local_path

        if tmpdir is not None:  # we're uploading to GCS or similar
            if process_should_upload:
                logger.info(f"Copying from temp dir to {path}")
                fs: AbstractFileSystem = fsspec.core.get_fs_token_paths(path, mode="wb")[0]
                fs.put(os.path.join(local_path, "*"), path, recursive=True)
                logger.info(f"Finished copying to {path}")
            else:
                logger.info(f"Waiting for process 0 to finish saving checkpoint to {path}")

            global _sync_count
            sync_global_devices(f"upload? {path}{_sync_count}")
            _sync_count += 1
    finally:
        if tmpdir is not None:
            shutil.rmtree(tmpdir, ignore_errors=True)



from typing import Optional

import jax
import numpy as np
from jax.sharding import Mesh


def local_device_grid_positions(mesh, process_index: Optional[int] = None) -> tuple[np.ndarray, np.ndarray]:
    """Returns a tuple of nd arrays, one for each axis, indicating the position of each device on the grid.
    Analogous to what np.where would return."""
    if process_index is None:
        process_index = jax.process_index()

    my_device_pos = np.vectorize(lambda dev: dev.process_index == process_index)(mesh.devices)
    return my_device_pos.nonzero()


def local_devices_mapping(mesh: Mesh, process_index: Optional[int] = None) -> dict[int, int]:
    """
    Handles the case when different devices in same process share the same data in TP.
    Returns a mapping from local devices' DP/FSDP group index in global mesh to local indices
    """
    local_device_pos = local_device_grid_positions(mesh, process_index)[:2]  # first 2 axes are DP axes.
    result = {}
    uid = 0
    for local_device_index in range(len(local_device_pos[0])):
        key = local_device_pos[0][local_device_index] * mesh.devices.shape[1] + local_device_pos[1][local_device_index]
        if key not in result:
            # when two devices maps to the same key (different TP index), they will get the same data
            result[key] = uid
            uid += 1
    return result


def process_mesh_mapping(mesh) -> dict[int, int]:
    """
    Handles the case when different processes share the same data in TP.
    If we envision each process as a subgrid of the mesh for its devices, this is the position of the process
    in the coarsened process-level mesh
    """
    devices = mesh.devices
    result = {}
    uid = 0
    leftmost2uid = {}
    # basic logic: process index -> upper-left device -> TP index 0 device -> process index -> uid
    for process_index in range(jax.process_count()):
        tmp = [np.min(axis) for axis in local_device_grid_positions(mesh, process_index)]
        tmp[-1] = 0  # we want the device with TP group index 0 in the same DP/FSDP group
        upper_left_position = tuple(tmp)  # in order to index into devices
        upper_left_process = devices[upper_left_position].process_index
        # assign uid to each process that has a device with TP group index 0
        if upper_left_process not in leftmost2uid:
            leftmost2uid[upper_left_process] = uid
            uid += 1
        this_uid = leftmost2uid[upper_left_process]
        result[process_index] = this_uid

    return result



import tempfile

import chex
import equinox as eqx
import numpy
import numpy as np
import pytest
import transformers
from jax import random

import haliax as hax

from levanter.models.attention import AttentionMask
from levanter.models.llama import LlamaAttention, LlamaConfig, LlamaDecoderLayer, LlamaLMHeadModel, LlamaRMSNorm
from levanter.models.rotary import DefaultRotaryEmbeddingsConfig, RotaryEmbeddings
from levanter.models.rotary import _rotate_half as levanter_rotate_half
from levanter.utils.jax_utils import parameter_count
from test_utils import check_load_config, check_model_works_with_seqlen, parameterize_with_configs, skip_if_no_torch


@skip_if_no_torch
def test_llama_config():
    # load HF config and convert to levanter config
    hf_config = transformers.LlamaConfig.from_pretrained("meta-llama/Llama-2-7b-hf")
    llama_config = LlamaConfig.from_hf_config(hf_config)

    # convert back to HF config
    config_overrides = {
        "_name_or_path": hf_config._name_or_path,
        "architectures": hf_config.architectures,
        "torch_dtype": hf_config.torch_dtype,
    }
    new_hf_config = llama_config.to_hf_config(
        vocab_size=hf_config.vocab_size,
        config_overrides=config_overrides,
    )

    # assert the content in new_hf_config is the same as hf_config
    for k in new_hf_config.__dict__.keys():
        if k in ["_commit_hash", "transformers_version"]:
            continue
        assert getattr(new_hf_config, k) == getattr(
            hf_config, k
        ), f"{k} {getattr(new_hf_config, k)} != {getattr(hf_config, k)}"


def test_llama_flops():
    # Check that the forward flops is within 10% of the naive calculation
    hf_config = transformers.LlamaConfig.from_pretrained("NousResearch/Llama-2-7b-hf")
    llama_config = LlamaConfig.from_hf_config(hf_config)
    n_params = 6.738415616e9
    ratio = llama_config.flops_per_token(hf_config.vocab_size) / (2 * n_params)
    assert ratio > 1.1, f"ratio {ratio} < 1.1"
    assert ratio < 1.2, f"ratio {ratio} > 1.2"


@skip_if_no_torch
def test_llama_rotary_embedding():
    import torch
    from transformers.models.llama.modeling_llama import LlamaRotaryEmbedding as HFLlamaRotaryEmbedding

    llama_config = _get_llama_config()
    HeadSize = llama_config.HeadSize
    Pos = llama_config.Pos
    hidden_dim = HeadSize.size
    seq_len = Pos.size
    key = random.PRNGKey(0)
    device = "cpu"

    x = random.normal(key, (1, seq_len))
    x_torch = torch.from_numpy(np.array(x))

    levanter_emb = DefaultRotaryEmbeddingsConfig().build(HeadSize=HeadSize, Pos=Pos)
    levanter_output = (levanter_emb.cos, levanter_emb.sin)

    hf_rope = HFLlamaRotaryEmbedding(dim=hidden_dim, max_position_embeddings=seq_len, device=device)
    hf_output = hf_rope(x_torch, torch.arange(seq_len).reshape(1, -1))

    for jax_out, torch_out in zip(levanter_output, hf_output):
        torch_out = torch_out.numpy()
        assert np.isclose(torch_out, np.array(jax_out.array), rtol=1e-2, atol=1e-2).all(), f"{torch_out} != {jax_out}"


@skip_if_no_torch
def test_apply_rotary_pos_emb():
    import torch
    from transformers.models.llama.modeling_llama import apply_rotary_pos_emb as hf_apply_rotary_pos_emb
    from transformers.models.llama.modeling_llama import rotate_half as hf_rotate_half

    def assert_equal_out(hax_out, torch_out: torch.Tensor):
        assert np.isclose(
            torch_out.numpy(), np.array(hax_out.array), rtol=1e-2, atol=1e-2
        ).all(), f"{torch_out} != {hax_out}"

    def named_array_to_tensor(named_array):
        return torch.from_numpy(np.array(named_array.array))

    llama_config = _get_llama_config()

    Pos = llama_config.Pos
    Heads = llama_config.Heads
    HeadSize = llama_config.HeadSize
    Batch = hax.Axis("batch", 2)

    # note here we switch Heads and Pos for the shape of the output tensors
    q = hax.random.normal(random.PRNGKey(0), (Batch, Pos, Heads, HeadSize))
    k = hax.random.normal(random.PRNGKey(1), (Batch, Pos, Heads, HeadSize))

    # Check the output of _rotate_half() from levanter and hf
    levanter_out_rf_q = levanter_rotate_half(q, HeadSize)
    levanter_out_rf_k = levanter_rotate_half(k, HeadSize)

    q_tensor = named_array_to_tensor(q).transpose(1, 2)  # needed for HF
    k_tensor = named_array_to_tensor(k).transpose(1, 2)
    hf_out_rf_q = hf_rotate_half(q_tensor).transpose(1, 2)  # re-transpose to match levanter
    hf_out_rf_k = hf_rotate_half(k_tensor).transpose(1, 2)

    assert_equal_out(levanter_out_rf_q, hf_out_rf_q)
    assert_equal_out(levanter_out_rf_k, hf_out_rf_k)

    # Check the output of _apply_rotary_pos_emb() from levanter and hf
    cos = hax.random.normal(random.PRNGKey(2), (Pos, HeadSize))
    sin = hax.random.normal(random.PRNGKey(3), (Pos, HeadSize))

    rot = RotaryEmbeddings(cos=cos, sin=sin)

    levanter_out_rope_q, levanter_out_rope_k = rot(HeadSize, q, k)
    cos_tensor = named_array_to_tensor(cos)[None, :, :]
    sin_tensor = named_array_to_tensor(sin)[None, :, :]

    hf_out_rope_q, hf_out_rope_k = hf_apply_rotary_pos_emb(q_tensor, k_tensor, cos_tensor, sin_tensor)
    hf_out_rope_q = hf_out_rope_q.transpose(1, 2)  # re-transpose to match levanter
    hf_out_rope_k = hf_out_rope_k.transpose(1, 2)
    assert_equal_out(levanter_out_rope_q, hf_out_rope_q)
    assert_equal_out(levanter_out_rope_k, hf_out_rope_k)


@skip_if_no_torch
@pytest.mark.parametrize("use_flash", [True, False])
@pytest.mark.parametrize("num_kv_heads", [1, 2, 4])
def test_llama_attention(use_flash, num_kv_heads):
    import torch
    from transformers.models.llama.modeling_llama import LlamaAttention as HFLlamaAttention

    config = _get_llama_config(use_flash=use_flash, num_kv_heads=num_kv_heads)

    attention = LlamaAttention.init(config=config, key=random.PRNGKey(0))

    state = hax.state_dict.to_torch_compatible_state_dict(attention)
    state = {k: torch.from_numpy(np.array(v)) for k, v in state.items()}
    hf_attention = HFLlamaAttention(config.to_hf_config(32000))
    hf_attention.load_state_dict(state, strict=True)

    x, mask = _get_random_inputs(config)
    x_torch = torch.from_numpy(np.array(x.array))
    batch_size = x_torch.shape[0]
    explicit_mask = torch.from_numpy(np.array(mask.materialize(config.Pos, config.KeyPos).array))
    mask_torch = explicit_mask.broadcast_to((batch_size, 1, -1, -1))

    # the torch mask is really a bias, so we need to invert it and make it a big negative number
    mask_torch = (mask_torch == 0).float() * -1e9

    out = attention(x, mask)
    position_ids = torch.arange(config.Pos.size).reshape(1, -1)
    hf_out = hf_attention(x_torch, position_ids=position_ids, attention_mask=mask_torch)

    # assert np.isclose(
    #     hf_out[0].detach().cpu().numpy(), np.array(out.array), rtol=1e-4, atol=1e-4
    # ).all(), f"{hf_out[0]} != {out}"
    chex.assert_trees_all_close(hf_out[0].detach().cpu().numpy(), out.array, rtol=1e-4, atol=1e-4)


def test_llama_param_counts_dont_change_with_seqlen():
    model = LlamaLMHeadModel.init(hax.Axis("v", 2048), _get_llama_config(seq_len=128), key=random.PRNGKey(0))
    model2 = LlamaLMHeadModel.init(hax.Axis("v", 2048), _get_llama_config(seq_len=256), key=random.PRNGKey(0))
    assert parameter_count(model) == parameter_count(model2)


@skip_if_no_torch
def test_llama_rms_norm():
    import torch
    from transformers.models.llama.modeling_llama import LlamaRMSNorm as HFLlamaRMSNorm

    config = _get_llama_config()
    ln = LlamaRMSNorm.init(config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias)
    hf_ln = HFLlamaRMSNorm(config.Embed.size, eps=config.layer_norm_epsilon)

    x, _ = _get_random_inputs(config)
    x_torch = torch.from_numpy(np.array(x.array))

    out = ln(x)
    hf_out = hf_ln(x_torch)

    assert np.isclose(
        hf_out.detach().cpu().numpy(), np.array(out.array), rtol=1e-6, atol=1e-6
    ).all(), f"{hf_out} != {out}"


@skip_if_no_torch
@pytest.mark.parametrize("num_kv_heads", [1, 2, 4])
def test_llama_decoder_layer(num_kv_heads):
    import torch
    from transformers.models.llama.modeling_llama import LlamaDecoderLayer as HFLlamaDecoderLayer

    llama_config = _get_llama_config(num_kv_heads=num_kv_heads)
    key = random.PRNGKey(0)
    llama_decoder_layer = LlamaDecoderLayer.init(config=llama_config, key=key)

    state = hax.state_dict.to_torch_compatible_state_dict(llama_decoder_layer)
    state = {k: torch.from_numpy(np.array(v)) for k, v in state.items()}
    hf_decoder_layer = HFLlamaDecoderLayer(llama_config.to_hf_config(32000), layer_idx=0)
    hf_decoder_layer.load_state_dict(state, strict=True)

    x, mask = _get_random_inputs(llama_config)
    x_torch = torch.from_numpy(np.array(x.array))
    batch_size = x_torch.shape[0]
    explicit_mask = torch.from_numpy(np.array(mask.materialize(llama_config.Pos, llama_config.KeyPos).array))
    mask_torch = explicit_mask.broadcast_to((batch_size, 1, -1, -1))
    mask_torch = (mask_torch == 0).float() * -1e9

    position_ids = torch.arange(llama_config.Pos.size).reshape(1, -1)

    out = llama_decoder_layer(x, mask)
    hf_out = hf_decoder_layer(x_torch, position_ids=position_ids, attention_mask=mask_torch)

    assert np.isclose(
        hf_out[0].detach().cpu().numpy(), np.array(out.array), rtol=1e-4, atol=1e-4
    ).all(), f"{hf_out[0]} != {out}"


@pytest.mark.parametrize("num_kv_heads", [1, 2, 4])
def test_llama_lm_head_model(num_kv_heads):
    llama_config = _get_llama_config(num_kv_heads=num_kv_heads)
    Batch = hax.Axis("batch", 2)
    Vocab = hax.Axis("vocab", 1000)
    Pos = llama_config.Pos
    input_ids = hax.random.randint(random.PRNGKey(0), (Batch, Pos), 0, Vocab.size)
    mask = AttentionMask.causal()

    llama_model = LlamaLMHeadModel.init(Vocab=Vocab, config=llama_config, key=random.PRNGKey(0))
    out = llama_model(input_ids, mask)
    assert out.array.shape == (Batch.size, Pos.size, Vocab.size)


@pytest.mark.parametrize("use_flash", [True, False])
@pytest.mark.parametrize("num_kv_heads", [1, 2, 4])
def test_llama_lm_head_model_bwd(use_flash, num_kv_heads):
    llama_config = _get_llama_config(use_flash=use_flash, num_kv_heads=num_kv_heads)
    Batch = hax.Axis("batch", 2)
    Vocab = hax.Axis("vocab", 1000)
    Pos = llama_config.Pos
    input_ids = hax.random.randint(random.PRNGKey(0), (Batch, Pos), 0, Vocab.size)
    mask = AttentionMask.causal()

    llama_model = LlamaLMHeadModel.init(Vocab=Vocab, config=llama_config, key=random.PRNGKey(0))

    def f(llama_model, input_ids, mask):
        out = llama_model(input_ids, mask)
        return hax.sum(out).scalar()

    _, grads = eqx.filter_value_and_grad(f)(llama_model, input_ids, mask)


@skip_if_no_torch
@pytest.mark.parametrize("scan_layers", [True, False])
@pytest.mark.parametrize("num_kv_heads", [1, 2, 4])
def test_llama_roundtrip(scan_layers, num_kv_heads):
    import torch
    from transformers import AutoModelForCausalLM, LlamaForCausalLM

    converter = LlamaConfig().hf_checkpoint_converter()

    config = LlamaConfig(
        seq_len=128,
        hidden_dim=16,
        num_heads=4,
        num_kv_heads=num_kv_heads,
        gradient_checkpointing=False,
        scan_layers=scan_layers,
    )
    Vocab = hax.Axis("vocab", 1000)
    hf_config = config.to_hf_config(Vocab.size)

    # Make input and attn_mask
    input = hax.random.randint(random.PRNGKey(0), config.Pos, 0, Vocab.size)
    attn_mask = AttentionMask.causal()
    input_torch = torch.from_numpy(np.array(input.array)).to(torch.int32).unsqueeze(0)

    torch.random.manual_seed(0)

    torch_model = LlamaForCausalLM(hf_config)
    torch_model.eval()

    torch_out = torch_model(input_torch)
    torch_out = torch_out.logits[0].detach().cpu().numpy()
    # torch_out = jax.nn.softmax(torch_out, axis=-1)

    with tempfile.TemporaryDirectory() as tmpdir:
        torch_model.save_pretrained(f"{tmpdir}/torch_model")

        model = converter.load_pretrained(
            LlamaLMHeadModel, ref=f"{tmpdir}/torch_model", resize_vocab_to_match_tokenizer=False
        )

        @hax.named_jit
        def compute(model, input):
            model_output = model(input, attn_mask=attn_mask)
            return model_output

        jax_out = compute(model, input).array

        assert torch_out.shape == jax_out.shape, f"{torch_out.shape} != {jax_out.shape}"
        assert np.isclose(torch_out, np.array(jax_out), rtol=1e-4, atol=1e-4).all(), f"{torch_out} != {jax_out}"

        # now we're going to magnify the model parameters enough that differences should actualy show up
        jax_out = compute(model, input).array

        converter.save_pretrained(model, f"{tmpdir}/lev_model", save_reference_code=False)
        torch_model2 = AutoModelForCausalLM.from_pretrained(f"{tmpdir}/lev_model")
        torch_model2.eval()

        torch_out2 = torch_model2(input_torch)
        torch_out2 = torch_out2.logits[0].detach().cpu().numpy()
        assert torch_out2.shape == jax_out.shape, f"{torch_out2.shape} != {jax_out.shape}"
        numpy.testing.assert_allclose(torch_out2, jax_out, rtol=1e-5, atol=1e-5)


def _get_llama_config(use_flash=False, num_kv_heads=4, seq_len=128) -> LlamaConfig:
    return LlamaConfig(
        seq_len=seq_len,
        hidden_dim=16,
        num_heads=4,
        num_kv_heads=num_kv_heads,
        gradient_checkpointing=False,  # disable for tests so debugging is easier
        use_flash_attention=use_flash,
        flash_attention_block_size=8 if use_flash else None,
    )


def _get_random_inputs(config: LlamaConfig):
    Embed = config.Embed
    Pos = config.Pos
    Batch = hax.Axis("batch", 2)
    x = hax.random.normal(random.PRNGKey(0), (Batch, Pos, Embed))
    mask = AttentionMask.causal()
    return x, mask


@parameterize_with_configs("llama*.yaml")
def test_llama_configs(config_file):
    from levanter.main.train_lm import TrainLmConfig

    config_class = TrainLmConfig

    check_load_config(config_class, config_file)


@pytest.mark.parametrize("num_kv_heads", [1, 2])
def test_pass_different_length_seq(num_kv_heads):
    config = LlamaConfig(
        seq_len=64,
        hidden_dim=64,
        intermediate_dim=32,
        num_heads=2,
        num_kv_heads=num_kv_heads,
        use_flash_attention=True,
    )
    check_model_works_with_seqlen(LlamaLMHeadModel, config, 16)


@skip_if_no_torch
@pytest.mark.parametrize("scan_layers", [True, False])
@pytest.mark.parametrize("num_kv_heads", [2, 4])
def test_state_dict_consistency(scan_layers, num_kv_heads):
    from transformers import LlamaForCausalLM

    config = LlamaConfig(
        seq_len=128,
        hidden_dim=16,
        num_heads=4,
        num_layers=4,
        num_kv_heads=num_kv_heads,
        gradient_checkpointing=False,
        scan_layers=scan_layers,
    )
    Vocab = hax.Axis("vocab", 1000)
    model = LlamaLMHeadModel.init(Vocab=Vocab, config=config, key=random.PRNGKey(0))
    hf_config = config.to_hf_config(Vocab.size)
    hf_model = LlamaForCausalLM(hf_config)
    levanter_state_dict = hax.state_dict.to_torch_compatible_state_dict(model)
    assert set(hf_model.state_dict().keys()) == set(levanter_state_dict.keys())



from levanter.utils.datetime_utils import encode_timedelta, parse_timedelta


def test_encode_timedelta():
    # various time strings from the unit tests for pytimeparse
    # https://github.com/wroberts/pytimeparse/blob/master/pytimeparse/tests/testtimeparse.py
    # i skipped negative ones because they're not supported, and fractional things because really

    def ensure_roundtrip(td_str, expected_seconds):
        # we don't enforce that the output is the same as the input,
        # but we do enforce that it can be parsed to the same timedelta
        td = parse_timedelta(td_str)
        assert td.total_seconds() == expected_seconds
        assert parse_timedelta(encode_timedelta(td)) == td, f"Failed to roundtrip {td_str}: {encode_timedelta(td)}"

    ensure_roundtrip("1d", 86400)
    ensure_roundtrip("+32 m 1 s", 1921)
    ensure_roundtrip("+ 32 m 1 s", 1921)
    ensure_roundtrip("32m", 1920)
    ensure_roundtrip("+32m", 1920)
    ensure_roundtrip("2h32m", 9120)
    ensure_roundtrip("+2h32m", 9120)
    ensure_roundtrip("3d2h32m", 268320)
    ensure_roundtrip("+3d2h32m", 268320)
    ensure_roundtrip("1w3d2h32m", 873120)
    ensure_roundtrip("1w 3d 2h 32m", 873120)
    ensure_roundtrip("1 w 3 d 2 h 32 m", 873120)
    ensure_roundtrip("4:13", 253)
    ensure_roundtrip(":13", 13)
    ensure_roundtrip("4:13:02", 15182)
    ensure_roundtrip("4:13:02.266", 15182.266)
    ensure_roundtrip("2:04:13:02.266", 187982.266)
    ensure_roundtrip("2 days,  4:13:02", 187982)
    ensure_roundtrip("5hr34m56s", 20096)
    ensure_roundtrip("5 hours, 34 minutes, 56 seconds", 20096)
    ensure_roundtrip("5 hrs, 34 mins, 56 secs", 20096)
    ensure_roundtrip("2 days, 5 hours, 34 minutes, 56 seconds", 192896)
    ensure_roundtrip("172 hr", 619200)



import dataclasses
from typing import TypeVar

import equinox as eqx
import jax
from jaxtyping import PyTree

from haliax.util import StringHolderEnum


T = TypeVar("T", bound=PyTree)


class NonePolicy(StringHolderEnum):
    PRESERVE = "preserve"
    REPLACE = "replace"
    ERROR = "error"


def inference_mode(tree: T, value: bool, none_policy: str = NonePolicy.REPLACE) -> T:
    """
    Analogous to [equinox.nn.inference_mode][] (neé [equinox.tree_inference][]), but
    it works in the presence of nones for the `inference` argument.
    """

    has_inference = lambda leaf: hasattr(leaf, "inference")  # noqa: E731

    def replace_fn(node):
        if not has_inference(node):
            return node

        if node.inference is None:
            if none_policy == NonePolicy.PRESERVE:
                return node
            elif none_policy == NonePolicy.ERROR:
                raise ValueError(f"None found in {tree}.inference with none_policy={none_policy}")
            else:
                assert none_policy == NonePolicy.REPLACE, f"Unknown none_policy {none_policy}"

        if dataclasses.is_dataclass(node):
            return dataclasses.replace(node, inference=value)
        else:
            return eqx.tree_at(lambda x: x.inference, node, value, is_leaf=lambda x: x is node)

    def rec_set(tree):
        if has_inference(tree):
            tree = replace_fn(tree)

        if jax.tree_util.tree_leaves(tree) == [tree]:
            return tree

        return jax.tree_util.tree_map(rec_set, tree, is_leaf=lambda x: has_inference(x) and tree is not x)

    return rec_set(tree)



from typing import Optional

import jax.lax
import jax.numpy as jnp

import haliax as hax
from haliax import Axis, NamedArray
from haliax.types import PrecisionLike


def causal_sliding_window_attention(
    Pos: Axis,
    Window: Axis,
    Head: Axis,
    query: NamedArray,
    key: NamedArray,
    value: NamedArray,
    bias: Optional[NamedArray] = None,
    attention_dtype: Optional[jnp.dtype] = None,
    precision: PrecisionLike = None,
) -> NamedArray:
    """
    Computes sliding window attention a la Longformer. This method uses blocking because Jax can't figure it
    out automatically.
    """
    # We use the window size as the block size
    # The basic idea is that we want to compute attention one block (of query) at a time, where a block is a window
    # of the sequence. Each q can attend to the prior window_size-1 positions plus itself
    assert Window.size <= Pos.size, "Window size must be at least 2x sequence length"
    assert Pos.size % Window.size == 0, "Sequence length must be divisible by window size"

    if Window.size == Pos.size:
        # we can just use regular attention
        # we have to special case this because jax won't like the attend_block_N function
        # which doesn't actually get executed but does get traced
        K = Pos.alias("K")
        return hax.nn.attention.dot_product_attention(
            K,
            Head,
            query,
            key.rename({Pos: K}),
            value.rename({Pos: K}),
            mask=hax.nn.attention.causal_mask(Pos, K),
            bias=bias,
            attention_dtype=attention_dtype,
            precision=precision,
        )

    # the attention structure is that each query attends to the prior window_size positions (q - window_size, q]
    # We extract one query block of length Window.size at a time (the block size and window size could be different, but
    # this seems fine)
    # For each query block, we extract a key and value block of length KWindow == Window.size * 2 - 1
    # The key block is [query_block_start - window_size + 1, query_block_start + window_size)

    # TODO: relax?
    Block = Axis("Block", Pos.size // Window.size)
    KWindow = Axis("KWindow", Window.size * 2 - 1)  # this is what we need to grab from the key/value

    # this makes code a bit easier to read below
    Q = Window
    K = KWindow

    # for our attention masks, each q can attend to the prior window_size-1 positions plus itself
    # that is, each q can attend to k s.t. k \in [q - window_size + 1, q]
    # however, note that K is offset by window_size - 1, so we need to shift the mask by that amount
    # this means that we want to mask out k s.t. k \in [q, q + window_size)
    # equivalently, k - q \in [0, window_size)
    diff = hax.arange(K) - hax.arange(Q).broadcast_axis(K)
    attn_mask = (diff >= 0) & (diff < Window.size)

    def attend_block_N(block_idx):
        block_idx = block_idx.scalar()
        query_block = query.slice(Pos, Q, start=block_idx * Q.size)
        # extract the relevant window from the key and value
        # this spans [query_block_start - window_size + 1, query_block_start + window_size)
        key_block = key.slice(Pos, K, start=(block_idx - 1) * Q.size + 1)
        value_block = value.slice(Pos, K, start=(block_idx - 1) * Q.size + 1)

        if bias is not None:
            bias_block = bias.slice(Pos, K, start=(block_idx - 1) * Q.size + 1)
        else:
            bias_block = None

        return hax.nn.attention.dot_product_attention(
            K, Head, query_block, key_block, value_block, attn_mask, bias_block, attention_dtype, precision
        )

    # for the 0th block, we have to worry about the out-of-bounds. just use a causal mask and do normal causal attention
    # NB if you change it so that the block size and window size aren't the same, you'll need to change this
    K0 = Q.alias("K0")
    attn_mask_0 = hax.nn.attention.causal_mask(Q, K0)

    def attend_block_0(block_idx):
        query_block = query.slice(Pos, Q, start=0)
        key_block = key.slice(Pos, K0, start=0)
        value_block = value.slice(Pos, K0, start=0)
        if bias is not None:
            bias_block = bias.slice(Pos, K0, start=0)
        else:
            bias_block = None
        return hax.nn.attention.dot_product_attention(
            K0, Head, query_block, key_block, value_block, attn_mask_0, bias_block, attention_dtype, precision
        )

    # extra arg/return for dummy scan accumulator
    def attend_block(_, block_idx):
        return None, jax.lax.cond(block_idx.scalar() == 0, attend_block_0, attend_block_N, block_idx)

    # we use scan here to encourage jax to do the blocking
    _, blocked_attn = hax.scan(attend_block, Block)(None, hax.arange(Block))  # type: ignore

    # now we need to unblock the attention
    # TODO: see if the rearrange and flatten_axes have perf implications
    return blocked_attn.flatten_axes((Block, Q), Pos).rearrange(value.axes)
